from py42._compat import str
from py42._compat import UserDict
from py42.clients.settings import SettingProperty
from py42.clients.settings import TSettingProperty
from py42.clients.settings._converters import bool_to_str
from py42.clients.settings._converters import bytes_to_gb
from py42.clients.settings._converters import comma_separated_to_list
from py42.clients.settings._converters import gb_to_bytes
from py42.clients.settings._converters import str_to_bool
from py42.clients.settings._converters import to_comma_separated
from py42.clients.settings._converters import to_list
from py42.clients.settings.device_settings import DeviceSettingsDefaults


class OrgSettings(UserDict, object):
    """Class used to manage an Organization's settings."""

    def __init__(self, org_settings, t_settings):
        self.data = org_settings
        self._t_settings = t_settings
        self._packets = {}
        self.changes = {}
        self.device_defaults = DeviceSettingsDefaults(
            self.data[u"deviceDefaults"], org_settings=self
        )

    @property
    def packets(self):
        """The setting packets for any modifications to be posted to the /api/OrgSettings
        endpoint.
        """
        return list(self._packets.values())

    @property
    def org_id(self):
        """The identifier for the org."""
        return self.data[u"orgId"]

    @property
    def registration_key(self):
        """The registration key for the org."""
        return self.data[u"registrationKey"]

    org_name = SettingProperty(u"org_name", [u"orgName"])
    """Name for this Org."""

    external_reference = SettingProperty(u"external_reference", [u"orgExtRef"])
    """External reference field for this Org."""

    notes = SettingProperty(u"notes", [u"notes"])
    """Notes field for this Org."""

    quota_settings_inherited = SettingProperty(
        u"quota_settings_inherited", [u"settings", u"isUsingQuotaDefaults"],
    )
    """Determines if Org Quota settings (`maximum_user_subscriptions`, `org_backup_quota`,
    `user_backup_quota`, `archive_hold_days`) are inherited from parent organization.

    Modifying one of the Org Quota attributes automatically sets this attribute to `False`.
    """

    archive_hold_days = SettingProperty(
        u"archive_hold_days",
        [u"settings", u"archiveHoldDays"],
        inheritance_attr=u"quota_settings_inherited",
    )
    """Number of days backup archives are held in cold storage after deactivation or
    destination removal from any devices in this Org.
    """

    maximum_user_subscriptions = SettingProperty(
        u"maximum_user_subscriptions",
        [u"settings", u"maxSeats"],
        inheritance_attr=u"quota_settings_inherited",
    )
    """Number of users allowed to consume a license in this Org. Set to -1 for unlimited."""

    org_backup_quota = SettingProperty(
        u"org_backup_quota",
        [u"settings", u"maxBytes"],
        get_converter=bytes_to_gb,
        set_converter=gb_to_bytes,
        inheritance_attr=u"quota_settings_inherited",
    )
    """Backup storage quota (in GB) for this organization. Set to -1 for unlimited."""

    user_backup_quota = SettingProperty(
        u"user_backup_quota",
        [u"settings", u"defaultUserMaxBytes"],
        get_converter=bytes_to_gb,
        set_converter=gb_to_bytes,
        inheritance_attr=u"quota_settings_inherited",
    )
    """Backup storage quota (in GB) for each user in this organization. Set to -1 for
    unlimited."""

    web_restore_admin_limit = SettingProperty(
        u"web_restore_admin_limit", [u"settings", u"webRestoreAdminLimitMb"]
    )
    """Limit (in MB) to amount of data restorable by admin users via web restore."""

    web_restore_user_limit = SettingProperty(
        u"web_restore_user_limit", [u"settings", u"webRestoreUserLimitMb"]
    )
    """Limit (in MB) to amount of data restorable by non-admin users via web restore."""

    reporting_settings_inherited = SettingProperty(
        u"reporting_settings_inherited", [u"settings", u"isUsingReportingDefaults"],
    )
    """Determines if Org Reporting settings (`backup_warning_email_days`,
    `backup_critical_email_days', `backup_alert_recipient_emails`) are inherited from
    parent organization.

    Modifying one of the Org Reporting attributes automatically sets this attribute to
    `False`.
    """

    backup_warning_email_days = SettingProperty(
        u"backup_warning_email_days",
        [u"settings", u"warnInDays"],
        inheritance_attr=u"reporting_settings_inherited",
    )
    """The number of days devices in this org can go without any backup before "warning"
    alerts get sent to org admins.
    """

    backup_critical_email_days = SettingProperty(
        u"backup_critical_email_days",
        [u"settings", u"alertInDays"],
        inheritance_attr=u"reporting_settings_inherited",
    )
    """The number of days devices in this org can go without any backup before "critical"
    alerts get sent to org admins.
    """

    backup_alert_recipient_emails = SettingProperty(
        u"backup_alert_recipient_emails",
        [u"settings", u"recipients"],
        set_converter=to_list,
        inheritance_attr=u"reporting_settings_inherited",
    )
    """List of email addresses that organization backup alert emails get sent to (org
    admin users get these automatically).
    """

    _endpoint_monitoring_enabled = TSettingProperty(
        u"endpoint_monitoring_enabled",
        u"org-securityTools-enable",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    _aed_enabled = TSettingProperty(
        u"aed_enabled",
        u"device_advancedExfiltrationDetection_enabled",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    _removable_media_enabled = TSettingProperty(
        u"removable_media_enabled",
        u"org-securityTools-device-detection-enable",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    _cloud_sync_enabled = TSettingProperty(
        u"cloud_sync_enabled",
        u"org-securityTools-cloud-detection-enable",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    _browser_and_applications_enabled = TSettingProperty(
        u"browser_and_applications_enabled",
        u"org-securityTools-open-file-detection-enable",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    _file_metadata_collection_enabled = TSettingProperty(
        u"file_metadata_collection_enabled",
        u"device_fileForensics_enabled",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    _printer_detection_enabled = TSettingProperty(
        u"printer_detection_enabled",
        u"org_securityTools_printer_detection_enable",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )

    @property
    def endpoint_monitoring_enabled(self):
        """Determines if endpoint monitoring settings are enabled for this org.

        Disabling this property also disables "removable media", "cloud sync",
        "browser and application monitoring" and "printer detection" properties.
        """
        return self._endpoint_monitoring_enabled

    @endpoint_monitoring_enabled.setter
    def endpoint_monitoring_enabled(self, val):
        self._endpoint_monitoring_enabled = val
        self._aed_enabled = val
        if not val:
            self._cloud_sync_enabled = val
            self._browser_and_applications_enabled = val
            self._removable_media_enabled = val
            self._printer_detection_enabled = val

    @property
    def endpoint_monitoring_removable_media_enabled(self):
        """Determines if removable media endpoint monitoring event capturing is enabled
        for this org.
        """
        return self._removable_media_enabled

    @endpoint_monitoring_removable_media_enabled.setter
    def endpoint_monitoring_removable_media_enabled(self, value):
        if value:
            self.endpoint_monitoring_enabled = value
        self._removable_media_enabled = value

    @property
    def endpoint_monitoring_cloud_sync_enabled(self):
        """Determines if cloud sync endpoint monitoring event capturing is enabled
        for this org.
        """
        return self._cloud_sync_enabled

    @endpoint_monitoring_cloud_sync_enabled.setter
    def endpoint_monitoring_cloud_sync_enabled(self, value):
        if value:
            self.endpoint_monitoring_enabled = value
        self._cloud_sync_enabled = value

    @property
    def endpoint_monitoring_browser_and_applications_enabled(self):
        """Determines if browser and other application activity endpoint monitoring
        event capturing is enabled for this org.
        """
        return self._browser_and_applications_enabled

    @endpoint_monitoring_browser_and_applications_enabled.setter
    def endpoint_monitoring_browser_and_applications_enabled(self, value):
        if value:
            self.endpoint_monitoring_enabled = value
        self._browser_and_applications_enabled = value

    @property
    def endpoint_monitoring_printer_detection_enabled(self):
        """Determines if printer endpoint monitoring event capturing is enabled for this
        org.
        """
        return self._printer_detection_enabled

    @endpoint_monitoring_printer_detection_enabled.setter
    def endpoint_monitoring_printer_detection_enabled(self, value):
        if value:
            self.endpoint_monitoring_enabled = value
        self._printer_detection_enabled = value

    @property
    def endpoint_monitoring_file_metadata_collection_enabled(self):
        """Determines if file metadata collection is enabled for this org."""
        return self._file_metadata_collection_enabled

    @endpoint_monitoring_file_metadata_collection_enabled.setter
    def endpoint_monitoring_file_metadata_collection_enabled(self, value):
        if value:
            self.endpoint_monitoring_enabled = value
        self._file_metadata_collection_enabled = value

    endpoint_monitoring_file_metadata_scan_enabled = TSettingProperty(
        u"file_metadata_scan_enabled",
        u"device_fileForensics_scan_enabled",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    """Determines if file metadata collection regular full scans are enabled for this
    org.
    """

    endpoint_monitoring_file_metadata_ingest_scan_enabled = TSettingProperty(
        u"file_metadata_ingest_scan_enabled",
        u"device_fileForensics_enqueue_scan_events_during_ingest",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    """Determines if file metadata collection does an initial full scan when first
    enabled on devices.
    """

    endpoint_monitoring_background_priority_enabled = TSettingProperty(
        u"background_priority_enabled",
        u"device_background_priority_enabled",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    """Determines if devices in this org have reduced priority in some IO bound tasks.
    If enabled, devices may see improved general device performance at the expense of
    some Code42 backup/security tasks taking longer.
    """

    endpoint_monitoring_custom_applications_win = TSettingProperty(
        u"custom_monitored_applications_win",
        u"device_org_winAppActivity_binaryWhitelist",
        get_converter=comma_separated_to_list,
        set_converter=to_comma_separated,
    )
    """List of additional applications the Code42 client monitors for file exfiltration
    activity.

    See `Support Documentation <https://support.code42.com/Administrator/Cloud/Configuring/Customize_applications_monitored_for_file_exfiltration>`__
    for more details.
    """

    endpoint_monitoring_custom_applications_mac = TSettingProperty(
        u"custom_monitored_applications_mac",
        u"device_org_macAppActivity_binaryWhitelist",
        get_converter=comma_separated_to_list,
        set_converter=to_comma_separated,
    )
    """List of additional applications the Code42 client monitors for file exfiltration
    activity.

    See `Support Documentation <https://support.code42.com/Administrator/Cloud/Configuring/Customize_applications_monitored_for_file_exfiltration>`__
    for more details.
    """

    endpoint_monitoring_file_metadata_collection_exclusions = TSettingProperty(
        u"file_metadata_collection_exclusions",
        u"device_fileForensics_fileExclusions_org",
    )
    """File types and file paths to exclude from file metadata collection.

    See `Support Documentation <https://support.code42.com/Administrator/Cloud/Configuring/File_Metadata_Collection_exclusions>`__
    for more details on the shape of the body this setting expects.
    """

    endpoint_monitoring_file_exfiltration_detection_exclusions = TSettingProperty(
        u"file_exfiltration_detection_exclusions",
        u"org_securityTools_detection_monitoring_exclusions",
    )
    """File types and file paths to exclude from file exfiltration detection.

    See `Support Documentation <https://support.code42.com/Administrator/Cloud/Configuring/Endpoint_monitoring#ExcludePaths>`__
    for more details on the shape of the body this setting expects.
    """

    web_restore_enabled = TSettingProperty(
        u"web_restore_enabled",
        u"device_webRestore_enabled",
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    """Determines if web restores are enabled for devices in this org."""

    def __repr__(self):
        return u"<OrgSettings: org_id: {}, name: '{}'>".format(
            self.data[u"orgId"], self.data[u"orgName"]
        )

    def __str__(self):
        return str(self.data)



from py42.services._auth import C42RenewableAuth
from py42.services._connection import Connection


class StorageAuth(C42RenewableAuth):
    def __init__(self):
        super(StorageAuth, self).__init__()
        self._storage_url = None
        self._storage_connection = None

    def get_storage_url(self):
        self.get_credentials()
        return self._storage_url

    def get_tmp_auth(self):
        raise NotImplementedError()

    def _get_auth_token(self, login_token):
        uri = u"api/AuthToken"
        response = self._storage_connection.post(
            uri, headers={u"Authorization": u"login_token {}".format(login_token)}
        )
        token1, token2 = response.data
        return u"token {}-{}".format(token1, token2)

    def _get_credentials(self):
        storage_url, login_token = self.get_tmp_auth()
        self._storage_url = storage_url
        self._storage_connection = (
            self._storage_connection or _get_new_storage_connection(self._storage_url)
        )
        return self._get_auth_token(login_token)


class FileArchiveAuth(StorageAuth):
    def __init__(self, connection, user_id, device_guid, destination_guid):
        super(FileArchiveAuth, self).__init__()
        self._connection = connection
        self._user_id = user_id
        self._device_guid = device_guid
        self._destination_guid = destination_guid

    def get_tmp_auth(self):
        uri = u"/api/LoginToken"
        data = {
            u"userId": self._user_id,
            u"sourceGuid": self._device_guid,
            u"destinationGuid": self._destination_guid,
        }
        response = self._connection.post(uri, json=data)
        storage_url = response[u"serverUrl"]
        login_token = response[u"loginToken"]
        return storage_url, login_token


class SecurityArchiveAuth(StorageAuth):
    def __init__(self, connection, plan_uid, destination_guid):
        super(SecurityArchiveAuth, self).__init__()
        self._connection = connection
        self._plan_uid = plan_uid
        self._destination_guid = destination_guid

    def get_tmp_auth(self):
        uri = u"/api/StorageAuthToken"
        data = {u"planUid": self._plan_uid, u"destinationGuid": self._destination_guid}
        response = self._connection.post(uri, json=data)
        storage_url = response[u"serverUrl"]
        login_token = response[u"loginToken"]
        return storage_url, login_token


def _get_new_storage_connection(storage_url):
    return Connection.from_host_address(storage_url)



from py42.sdk.queries.fileevents.file_event_query import FileEventFilterComparableField
from py42.sdk.queries.fileevents.file_event_query import FileEventFilterStringField
from py42.util import get_attribute_keys_from_class


class FileCategory(FileEventFilterStringField):
    """Class that filters events by category of the file observed.

    Available file categories are provided as class attributes:
        - :attr:`FileCategory.AUDIO`
        - :attr:`FileCategory.DOCUMENT`
        - :attr:`FileCategory.EXECUTABLE`
        - :attr:`FileCategory.IMAGE`
        - :attr:`FileCategory.PDF`
        - :attr:`FileCategory.PRESENTATION`
        - :attr:`FileCategory.SCRIPT`
        - :attr:`FileCategory.SOURCE_CODE`
        - :attr:`FileCategory.SPREADSHEET`
        - :attr:`FileCategory.VIDEO`
        - :attr:`FileCategory.VIRTUAL_DISK_IMAGE`
        - :attr:`FileCategory.ZIP`

    """

    _term = u"fileCategory"

    AUDIO = u"Audio"
    DOCUMENT = u"Document"
    EXECUTABLE = u"Executable"
    IMAGE = u"Image"
    PDF = u"Pdf"
    PRESENTATION = u"Presentation"
    SCRIPT = u"Script"
    SOURCE_CODE = u"SourceCode"
    SPREADSHEET = u"Spreadsheet"
    VIDEO = u"Video"
    VIRTUAL_DISK_IMAGE = u"VirtualDiskImage"
    ZIP = u"Archive"

    @staticmethod
    def choices():
        return get_attribute_keys_from_class(FileCategory)


class FileName(FileEventFilterStringField):
    """Class that filters events by the name of the file observed."""

    _term = u"fileName"


class FileOwner(FileEventFilterStringField):
    """Class that filters events by the owner of the file observed."""

    _term = u"fileOwner"


class FilePath(FileEventFilterStringField):
    """Class that filters events by path of the file observed."""

    _term = u"filePath"


class FileSize(FileEventFilterComparableField):
    """Class that filters events by size of the file observed.

    Size ``value`` must be bytes.
    """

    _term = u"fileSize"


class MD5(FileEventFilterStringField):
    """Class that filters events by the MD5 hash of the file observed."""

    _term = u"md5Checksum"


class SHA256(FileEventFilterStringField):
    """Class that filters events by SHA256 hash of the file observed."""

    _term = u"sha256Checksum"



from py42._compat import str
from py42._compat import UserDict
from py42._compat import UserList
from py42.clients.settings import check_lock
from py42.clients.settings import SettingProperty
from py42.clients.settings import show_change
from py42.clients.settings._converters import bool_to_str
from py42.clients.settings._converters import days_to_minutes
from py42.clients.settings._converters import minutes_to_days
from py42.clients.settings._converters import str_to_bool
from py42.exceptions import Py42Error

invalid_destination_error = Py42Error(
    u"Invalid destination guid or destination not offered to device's Org."
)
destination_not_added_error = Py42Error(
    u"Destination is not added to device, unable to lock."
)


class DeviceSettingsDefaults(UserDict, object):
    """Class used for managing an Organization's Device Default settings. Also acts as a
    base class for `DeviceSettings` to manage individual device settings."""

    def __init__(self, device_dict, org_settings):
        self.data = device_dict
        self._org_settings = org_settings
        self.changes = org_settings.changes
        self._destinations = org_settings.data[u"settings"][u"destinations"]
        self.data[u"settings"] = {
            u"serviceBackupConfig": self.data[u"serviceBackupConfig"]
        }
        bs = self.data[u"serviceBackupConfig"][u"backupConfig"][u"backupSets"]
        self.backup_sets = self._extract_backup_sets(bs)

    def _extract_backup_sets(self, backup_sets):
        if isinstance(backup_sets, dict):  # number of sets are locked
            backup_sets = backup_sets["backupSet"]
            if isinstance(backup_sets, dict):  # there's only one set configured
                return [BackupSet(self, backup_sets)]
            elif isinstance(backup_sets, list):
                return [
                    BackupSet(self, bs)
                    for bs in backup_sets
                    if not _backup_set_is_legal_hold(bs)
                ]
            else:
                raise Py42Error("Unable to extract backup sets: {}".format(backup_sets))
        else:
            return [
                BackupSet(self, bs)
                for bs in backup_sets
                if not _backup_set_is_legal_hold(bs)
            ]

    @property
    def available_destinations(self):
        """Returns a dict of destinations available to be used by devices. Dict keys are
        destination guids and values are destination names.
        """
        return {d[u"guid"]: d[u"destinationName"] for d in self._destinations}

    warning_email_enabled = SettingProperty(
        name=u"warning_email_enabled",
        location=[u"settings", u"serviceBackupConfig", u"warningEmailEnabled"],
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    """Determines if backup "warning" threshold email alerts are configured for this device."""

    critical_email_enabled = SettingProperty(
        name=u"critical_email_enabled",
        location=[u"settings", u"serviceBackupConfig", u"severeEmailEnabled"],
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    """Determines if backup "critical" threshold email alerts are configured for this device."""

    warning_alert_days = SettingProperty(
        name=u"warning_alert_days",
        location=[u"settings", u"serviceBackupConfig", u"minutesUntilWarning"],
        get_converter=minutes_to_days,
        set_converter=days_to_minutes,
    )
    """The number of days a device can go without any backup activity before
    "warning" alert threshold is passed.
    """

    critical_alert_days = SettingProperty(
        name=u"critical_alert_days",
        location=[u"settings", u"serviceBackupConfig", u"minutesUntilSevere"],
        get_converter=minutes_to_days,
        set_converter=days_to_minutes,
    )
    """The number of days a device can go without any backup activity before "warning"
    alert threshold is passed.
    """

    backup_status_email_enabled = SettingProperty(
        name=u"backup_status_email_enabled",
        location=[u"settings", u"serviceBackupConfig", u"backupStatusEmailEnabled"],
        get_converter=str_to_bool,
        set_converter=bool_to_str,
    )
    """Determines if the regularly scheduled backup status email is enabled."""

    backup_status_email_frequency_days = SettingProperty(
        name=u"backup_status_email_frequency_days",
        location=[
            u"settings",
            u"serviceBackupConfig",
            u"backupStatusEmailFreqInMinutes",
        ],
        get_converter=minutes_to_days,
        set_converter=days_to_minutes,
    )
    """Determines the frequency of the regularly scheduled backup status email."""

    def __repr__(self):
        return u"<DeviceSettingsDefaults: org_id: {}>".format(self._org_settings.org_id)


class DeviceSettings(DeviceSettingsDefaults):
    """Class used to manage an individual device's settings."""

    def __init__(self, device_dict):
        self.changes = {}
        self.data = device_dict
        self._destinations = device_dict[u"availableDestinations"]
        bs = self.data[u"settings"][u"serviceBackupConfig"][u"backupConfig"][
            u"backupSets"
        ]
        self.backup_sets = self._extract_backup_sets(bs)
        """List of :class:`BackupSet` objects used to manage this device's backup set configurations."""

    @property
    def computer_id(self):
        """Identifier of this device. Read-only."""
        return self.data[u"computerId"]

    @property
    def device_id(self):
        """Identifier of this device (alias of `.computer_id`). Read only."""
        return self.computer_id

    @property
    def guid(self):
        """Globally unique identifier of this device. Read-only."""
        return self.data[u"guid"]

    @property
    def org_id(self):
        """Identifier of the organization this device belongs to. Read-only."""
        return self.data[u"orgId"]

    @property
    def user_id(self):
        """Identifier of the user this device belongs to. Read-only."""
        return self.data[u"userId"]

    @property
    def version(self):
        """Latest reported Code42 client version number for this device. Read-only."""
        return self.data[u"version"]

    @property
    def java_memory_heap_max(self):
        """The maximum memory the client will use on its system"""
        return self.data[u"settings"][u"javaMemoryHeapMax"]

    name = SettingProperty(name=u"name", location=[u"name"])
    """Name for this device."""

    external_reference = SettingProperty(
        name=u"external_reference", location=[u"computerExtRef"]
    )
    """External reference field for this device."""

    notes = SettingProperty(name=u"notes", location=[u"notes"])
    """Notes field for this device."""

    def __repr__(self):
        return u"<DeviceSettings: guid: {}, name: {}>".format(
            self.data[u"guid"], self.data[u"name"]
        )


class BackupSet(UserDict, object):
    """Helper class for managing device backup sets and Org device default backup sets."""

    def __init__(self, settings_manager, backup_set_dict):
        self._manager = settings_manager
        self._changes = settings_manager.changes
        self.data = backup_set_dict
        includes, excludes = self._extract_file_selection_lists()
        regex_excludes = self._extract_regex_exclusions()
        self._included_files = TrackedFileSelectionList(
            self, u"included_files", includes, self._changes
        )
        self._excluded_files = TrackedFileSelectionList(
            self, u"excluded_files", excludes, self._changes
        )
        self._filename_exclusions = TrackedFileSelectionList(
            self, u"filename_exclusions", regex_excludes, self._changes
        )
        self._orig_destinations = self.destinations

    def _extract_file_selection_lists(self):
        """Converts the file selection portion of the settings dict ("pathset") into two
        lists of just paths, `included` and `excluded`.

        The "pathset" object is a different shape depending on how many paths it
        contains and whether its locked or not:
            No paths:           `[{"@cleared": "true", "@os": "Linux"}]`
            No paths locked:    `{'@locked': 'true', 'paths': {'@cleared': 'true', '@os': 'Linux'}}`
            One path:           `[{"path": {"@include": "C:/"}, "@os": "Linux"}]`
            One path locked:    `{'@locked': 'true', 'paths': {'@os': 'Linux', 'path': {'@include': 'C:/'}}}`
            One+ paths:         `[{"path": [{"@include": "C:/Users/"},{"@exclude": "C:/Users/Admin/"},],"@os": "Linux"}]`
            One+ paths locked:  `{'@locked': 'true', 'paths': {'@os': 'Linux', 'path': [{'@include': 'C:/Users/'}, {'@exclude': 'C:/Users/Admin/'}]}}`
        """
        pathset = self.data[u"backupPaths"][u"pathset"]
        if isinstance(pathset, dict):  # pathset is locked
            path_list = pathset[u"paths"].get(u"path")
        else:
            path_list = pathset[0].get(u"path")

        # no paths selected
        if path_list is None:
            return [], []

        # one path selected
        if isinstance(path_list, dict):
            path_list = [path_list]

        includes = [p[u"@include"] for p in path_list if u"@include" in p]
        excludes = [p[u"@exclude"] for p in path_list if u"@exclude" in p]
        return includes, excludes

    def _extract_regex_exclusions(self):
        """Converts the filename exclusion portion ("excludeUser") of the settings dict
        into a simple list of regex patterns.

        The "excludeUser" object is a different shape based on the number of exclusion
        patterns present and whether the setting is locked or not:
            No exclusions:          `[{"windows": [], "linux": [], "macintosh": []}]`
            No exclusions locked:   `{'@locked': 'true', 'patternList': {'windows': [], 'macintosh': [], 'linux': []}}`
            One exclusion:          `[{"windows": [], "pattern": {"@regex": ".*"}, "linux": [], "macintosh": []}]`
            One exclusion locked:   `{'@locked': 'true', 'patternList': {'pattern': {'@regex': '.*'}, 'windows': [], 'macintosh': [], 'linux': []}}`
            One+ exclusions:        `[{"windows": [], "pattern": [{"@regex": ".*1"}, {"@regex": ".*2"}],"linux": [],"macintosh": []}]
            One+ exclusion locked:  `{'@locked': 'true', 'patternList': {'pattern': [{'@regex': '.*1'}, {'@regex': '.*2'}], 'windows': [], 'macintosh': [], 'linux': []}}`
        """
        exclude_user = self.data[u"backupPaths"][u"excludeUser"]
        if isinstance(exclude_user, dict):  # exclusions are locked
            pattern_list = exclude_user[u"patternList"].get(u"pattern")
        else:
            pattern_list = exclude_user[0].get(u"pattern")
        if not pattern_list:
            return []
        if isinstance(pattern_list, dict):
            pattern_list = [pattern_list]
        return [p[u"@regex"] for p in pattern_list]

    def _build_file_selection(self):
        """Converts the user-friendly lists of included and excluded file paths back
        into a "pathset" object the api expects. Called whenever one of the file selection
        property lists (`.included_files`, `.excluded_files`) is modified.
        """
        paths = {u"@os": u"Linux", u"path": []}
        if not self._included_files:  # ignore excluded values if nothing is included
            paths[u"@cleared"] = u"true"
        else:
            path_list = []
            for path in self._included_files:
                path_list.append({u"@include": path, u"@und": u"false"})
            for path in self._excluded_files:
                path_list.append({u"@exclude": path, u"@und": u"false"})
            paths[u"path"] = path_list
            paths[u"@cleared"] = u"false"

        self.data[u"backupPaths"][u"pathset"] = {u"paths": paths}

    def _build_regex_exclusions(self):
        """Converts the user-friendly list of filename exclusions back into the
        "excludeUser" object the api expects. Called whenever the `.filename_exclusions`
        property list is modified.
        """
        patterns = []
        for regex in self._filename_exclusions:
            patterns.append({u"@regex": regex})
        user_exclude_dict = {
            u"patternList": {
                u"pattern": patterns,
                u"windows": {u"pattern": []},
                u"macintosh": {u"pattern": []},
                u"linux": {u"pattern": []},
            }
        }
        self.data[u"backupPaths"][u"excludeUser"] = user_exclude_dict

    @property
    def locked(self):
        """Indicates whether the backup set as a whole is locked. If True, individual
        settings for this backup set (except for Destination settings), cannot be modified.
        """
        return u"@locked" in self.data and str_to_bool(self.data[u"@locked"])

    @property
    def included_files(self):
        """Returns the list of files/folders included in the backup selection. Items can
        be added/removed from this list via normal list methods, or assigning a new list
        of files to this attribute to replace the existing one.
        """
        return self._included_files

    @included_files.setter
    def included_files(self, value):
        if isinstance(value, (list, tuple)):
            self._included_files.clear()
            self._included_files.extend(value)
        else:
            raise AttributeError(u"included files must be a list/tuple.")

    @property
    def excluded_files(self):
        """Returns the list of files/folders excluded from the backup selection. Items can
        be added/removed from this list via normal list methods, or assigning a new list
        of files to this attribute to replace the existing one.
        """
        return self._excluded_files

    @excluded_files.setter
    def excluded_files(self, value):
        if isinstance(value, (list, tuple)):
            self._excluded_files.clear()
            self._excluded_files.extend(value)
        else:
            raise AttributeError(u"excluded files must be a list/tuple.")

    @property
    def filename_exclusions(self):
        """Returns the list of regex patterns used to exclude file paths from the backup
        selection. Items can be added/removed from this list via normal list methods,
        or assigning a new list of patterns to this attribute to replace the existing
        one.
        """
        return self._filename_exclusions

    @filename_exclusions.setter
    def filename_exclusions(self, value):
        if isinstance(value, (list, tuple)):
            self._filename_exclusions.clear()
            self._filename_exclusions.extend(value)
        else:
            raise AttributeError(u"filename exclusions must be a list/tuple.")

    @property
    def destinations(self):
        """Returns a dict of the destinations used for backup for the backup set. Dict
        keys are the destination guids, values are the destination names.
        """
        destination_dict = {}
        if u"@cleared" in self.data[u"destinations"]:
            return destination_dict
        for d in self.data[u"destinations"]:
            guid = d[u"@id"]
            dest_name = self._manager.available_destinations[guid]
            if u"@locked" in d:
                dest_name = dest_name + u" <LOCKED>"
            destination_dict[guid] = dest_name
        return destination_dict

    def add_destination(self, destination_guid):
        """Adds a destination to be used by this backup set. Raises a :class:`Py42Error` if
        the supplied destination guid is not available to the parent device/org.

        Args:
            destination_guid (str, int): The globally unique identifier of the
                destination to be added.
        """
        destination_guid = str(destination_guid)
        if destination_guid in self._manager.available_destinations:
            if destination_guid not in self.destinations:
                if not self.destinations:  # no destinations
                    self.data[u"destinations"] = [{u"@id": destination_guid}]
                else:
                    self.data[u"destinations"].append({u"@id": destination_guid})
                self._changes[u"destinations"] = show_change(
                    self._orig_destinations, self.destinations
                )
        else:
            raise invalid_destination_error

    def remove_destination(self, destination_guid):
        """Removes a destination from use by this backup set.

        Args:
            destination_guid (str, int): The globally unique identifier of the
                destination to be removed.
        """
        destination_guid = str(destination_guid)
        self._raise_if_invalid_destination(destination_guid)
        if destination_guid in self.destinations:
            for d in self.data[u"destinations"]:
                if d[u"@id"] == destination_guid:
                    self.data[u"destinations"].remove(d)
            if not self.data[u"destinations"]:  # all destinations removed
                self.data[u"destinations"] = {u"@cleared": u"true"}
            self._changes[u"destinations"] = show_change(
                self._orig_destinations, self.destinations
            )

    def lock_destination(self, destination_guid):
        """Locks an in-use destination, disallowing the device owner from removing this
        destination from their backup. Raises a :class:`Py42Error` if the supplied destination
        guid is not in use on this backup set, or not available to the parent device/org.
        """
        destination_guid = str(destination_guid)
        if destination_guid in self._manager.available_destinations:
            if destination_guid not in self.destinations:
                raise destination_not_added_error
            else:
                for d in self.data[u"destinations"]:
                    if d[u"@id"] == destination_guid:
                        d[u"@locked"] = u"true"
                self._changes[u"destinations"] = show_change(
                    self._orig_destinations, self.destinations
                )
        else:
            raise invalid_destination_error

    def unlock_destination(self, destination_guid):
        """Unlocks an in-use destination, allowing the device owner to remove this
        destination from their backup. Raises a :class:`Py42Error` if the supplied destination
        guid is not in use on this backup set, or not available to the parent device/org.
        """
        destination_guid = str(destination_guid)
        self._raise_if_invalid_destination(destination_guid)
        if destination_guid not in self.destinations:
            raise destination_not_added_error
        else:
            for d in self.data[u"destinations"]:
                if d[u"@id"] == destination_guid:
                    del d[u"@locked"]
            self._changes[u"destinations"] = show_change(
                self._orig_destinations, self.destinations
            )

    def _raise_if_invalid_destination(self, destination_guid):
        if destination_guid not in self._manager.available_destinations:
            raise invalid_destination_error

    def __repr__(self):
        if isinstance(self.data[u"name"], dict):  # name setting locked
            name = self.data[u"name"][u"#text"]
        else:
            name = self.data[u"name"]
        return u"<BackupSet: id: {}, name: '{}'>".format(self.data[u"@id"], name)

    def __str__(self):
        return str(dict(self))


class TrackedFileSelectionList(UserList, object):
    """Helper class to track modifications to file selection lists."""

    def __init__(self, backup_set, name, _list, changes_dict):
        self.backup_set = backup_set
        self.name = name
        self.orig = list(_list)
        self.data = _list
        self._changes = changes_dict

    def register_change(self):
        self.backup_set._build_file_selection()
        self.backup_set._build_regex_exclusions()
        if set(self.orig) != set(self.data):
            self._changes[self.name] = show_change(self.orig, self.data)
        elif self.name in self._changes:
            del self._changes[self.name]

    @check_lock("backup_set")
    def append(self, item):
        self.data.append(item)
        self.register_change()

    @check_lock("backup_set")
    def clear(self):
        self.data.clear()
        self.register_change()

    @check_lock("backup_set")
    def extend(self, other):
        self.data.extend(other)
        self.register_change()

    @check_lock("backup_set")
    def insert(self, i, item):
        self.data.insert(i, item)
        self.register_change()

    @check_lock("backup_set")
    def pop(self, index=-1):
        value = self.data.pop(index)
        self.register_change()
        return value

    @check_lock("backup_set")
    def remove(self, value):
        self.data.remove(value)
        self.register_change()


def _backup_set_is_legal_hold(bs):
    """A legal hold backup set has 'destinations' locked since it enforces backup to
    _all_ offered destinations for the legal hold set. This transforms the parent
    'destinations' object from a list to a dict and adds the '@locked':'true' key/value
    pair, then the actual destinations list is under a new 'destination' key.

    '@locked' also has to be present as 'destinations' also can be a dict if _no_
    destinations have been added to the set; it becomes {'@cleared': 'true'}.
    """
    return isinstance(bs["destinations"], dict) and "@locked" in bs["destinations"]



from datetime import datetime

from py42.exceptions import Py42BadRequestError
from py42.exceptions import Py42NotFoundError
from py42.exceptions import Py42UserNotOnListError
from py42.services import BaseService
from py42.services.detectionlists import _DetectionListFilters
from py42.services.detectionlists import _PAGE_SIZE
from py42.services.detectionlists import handle_user_already_added_error
from py42.services.util import get_all_pages
from py42.util import get_attribute_keys_from_class

_DATE_FORMAT = "%Y-%m-%d"


class DepartingEmployeeFilters(_DetectionListFilters):
    """Constants available for filtering Departing Employee search results."""

    LEAVING_TODAY = u"LEAVING_TODAY"

    @staticmethod
    def choices():
        return get_attribute_keys_from_class(DepartingEmployeeFilters)


class DepartingEmployeeService(BaseService):
    """A service for interacting with Code42 Departing Employee APIs."""

    _uri_prefix = u"v2/departingemployee/{0}"

    _CREATED_AT = u"CREATED_AT"

    def __init__(self, session, user_context, user_profile_service):
        super(DepartingEmployeeService, self).__init__(session)
        self._user_context = user_context
        self._user_profile_service = user_profile_service

    def add(self, user_id, departure_date=None):
        """Adds a user to the Departing Employees list.
        `REST Documentation <https://developer.code42.com/api/#operation/DepartingEmployeeControllerV2_AddEmployee>`__

        Raises a :class:`Py42UserAlreadyAddedError` when a user already exists in the Departing Employee \
            detection list.

        Args:
            user_id (str or int): The Code42 userUid of the user you want to add to the departing \
                employees list.
            departure_date (str or datetime, optional): Date in yyyy-MM-dd format or instance of datetime.
                Date is treated as UTC. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        if isinstance(departure_date, datetime):
            departure_date = departure_date.strftime(_DATE_FORMAT)
        tenant_id = self._user_context.get_current_tenant_id()
        data = {
            u"tenantId": tenant_id,
            u"userId": user_id,
            u"departureDate": departure_date,
        }
        uri = self._uri_prefix.format(u"add")
        try:
            return self._connection.post(uri, json=data)
        except Py42BadRequestError as err:
            handle_user_already_added_error(err, user_id, u"departing-employee list")
            raise

    def remove(self, user_id):
        """Removes a user from the Departing Employees list.
        `REST Documentation <https://developer.code42.com/api/#operation/DepartingEmployeeControllerV2_RemoveUser>`__

        Args:
            user_id (str or int): The Code42 userUid of the user.

        Returns:
            :class:`py42.response.Py42Response`
        """

        tenant_id = self._user_context.get_current_tenant_id()
        uri = self._uri_prefix.format(u"remove")
        data = {u"userId": user_id, u"tenantId": tenant_id}
        try:
            return self._connection.post(uri, json=data)
        except Py42NotFoundError as err:
            raise Py42UserNotOnListError(err, user_id, u"departing-employee")

    def get(self, user_id):
        """Gets departing employee data of a user.
        `REST Documentation <https://developer.code42.com/api/#operation/DepartingEmployeeControllerV2_GetEmployee>`__

        Args:
            user_id (str or int): The Code42 userUid of the user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        tenant_id = self._user_context.get_current_tenant_id()
        uri = self._uri_prefix.format(u"get")
        data = {u"userId": user_id, u"tenantId": tenant_id}
        return self._connection.post(uri, json=data)

    def get_all(
        self,
        filter_type=DepartingEmployeeFilters.OPEN,
        sort_key=_CREATED_AT,
        sort_direction=u"DESC",
        page_size=_PAGE_SIZE,
    ):
        """Gets all Departing Employees.

        Args:
            filter_type (str, optional): ``EXFILTRATION_30_DAYS``, ``EXFILTRATION_24_HOURS``,
                ``OPEN``, or ``LEAVING_TODAY``. Constants are available at
                :class:`py42.services.detectionlists.departing_employee.DepartingEmployeeFilters`.
                Defaults to "OPEN".
            sort_key (str, optional): Sort results based by field. Defaults to "CREATED_AT".
            sort_direction (str, optional): ``ASC`` or ``DESC``. Defaults to "DESC".
            page_size (int, optional): The number of departing employees to return
                per page. Defaults to 100.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of departing employees.
        """
        return get_all_pages(
            self.get_page,
            u"items",
            filter_type=filter_type,
            sort_key=sort_key,
            sort_direction=sort_direction,
            page_size=page_size or _PAGE_SIZE,
        )

    def get_page(
        self,
        page_num,
        filter_type=DepartingEmployeeFilters.OPEN,
        sort_key=_CREATED_AT,
        sort_direction=u"DESC",
        page_size=_PAGE_SIZE,
    ):
        """Gets a single page of Departing Employees.

        Args:
            page_num (int): The page number to request.
            filter_type (str, optional): ``EXFILTRATION_30_DAYS``, ``EXFILTRATION_24_HOURS``,
                ``OPEN``, or ``LEAVING_TODAY``. Constants are available at
                :class:`py42.services.detectionlists.departing_employee.DepartingEmployeeFilters`.
                Defaults to "OPEN".
            sort_key (str, optional): Sort results based by field. Defaults to "CREATED_AT".
            sort_direction (str. optional): ``ASC`` or ``DESC``. Defaults to "DESC".
            page_size (int, optional): The number of departing employees to return
                per page. Defaults to 100.

        Returns:
            :class:`py42.response.Py42Response`
        """

        uri = self._uri_prefix.format(u"search")
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"pgSize": page_size,
            u"pgNum": page_num,
            u"filterType": filter_type,
            u"srtKey": sort_key,
            u"srtDirection": sort_direction,
        }
        return self._connection.post(uri, json=data)

    def set_alerts_enabled(self, alerts_enabled=True):
        """Enable or disable email alerting on Departing Employee exposure events.
        `REST Documentation <https://developer.code42.com/api/#operation/DepartingEmployeeControllerV2_SetAlertState>`__

        Args:
            alerts_enabled (bool): Set alerting to on (True) or off (False). Defaults to True.

        Returns:
            :class:`py42.response.Py42Response`
        """
        tenant_id = self._user_context.get_current_tenant_id()
        uri = self._uri_prefix.format(u"setalertstate")
        data = {u"tenantId": tenant_id, u"alertsEnabled": alerts_enabled}
        return self._connection.post(uri, json=data)

    def update_departure_date(self, user_id, departure_date):
        """Add or modify details of an existing Departing Employee case.
        `REST Documentation <https://developer.code42.com/api/#operation/DepartingEmployeeControllerV2_UpdateDepartureDate>`__

        Args:
            user_id (str): The Code42 userUid of the user.
            departure_date (str or datetime): Date in yyyy-MM-dd format or instance of datetime.
                Date is treated as UTC.

        Returns:
            :class:`py42.response.Py42Response`
        """

        tenant_id = self._user_context.get_current_tenant_id()
        if isinstance(departure_date, datetime):
            departure_date = departure_date.strftime(_DATE_FORMAT)
        uri = self._uri_prefix.format(u"update")
        data = {
            u"tenantId": tenant_id,
            u"userId": user_id,
            u"departureDate": departure_date,
        }
        return self._connection.post(uri, json=data)



from py42.sdk.queries.fileevents.file_event_query import FileEventFilterStringField
from py42.sdk.queries.query_filter import QueryFilterBooleanField
from py42.util import get_attribute_keys_from_class


class Actor(FileEventFilterStringField):
    """Class that filters events by the cloud service username of the event originator
    (applies to cloud data source events only).
    """

    _term = u"actor"


class DirectoryID(FileEventFilterStringField):
    """Class that filters events by unique identifier of the cloud drive or folder where the event
     occurred (applies to cloud data source events only).
    """

    _term = u"directoryId"


class Shared(QueryFilterBooleanField):
    """Class that filters events by the shared status of the file at the time the event occurred
    (applies to cloud data source events only).
    """

    _term = u"shared"


class SharedWith(FileEventFilterStringField):
    """Class that filters events by the list of users who had been granted access to the file at the
    time of the event (applies to cloud data source events only).
    """

    _term = u"sharedWith"


class SharingTypeAdded(FileEventFilterStringField):
    """Class that filters results to include events where a file's sharing permissions were
    changed to a value that increases exposure (applies to cloud data source events only).

    Available options provided as class attributes:
        - :attr:`SharingTypeAdded.SHARED_VIA_LINK`
        - :attr:`SharingTypeAdded.IS_PUBLIC`
        - :attr:`SharingTypeAdded.OUTSIDE_TRUSTED_DOMAIN`
    """

    _term = u"sharingTypeAdded"

    SHARED_VIA_LINK = u"SharedViaLink"
    IS_PUBLIC = u"IsPublic"
    OUTSIDE_TRUSTED_DOMAIN = u"OutsideTrustedDomains"

    @staticmethod
    def choices():
        return get_attribute_keys_from_class(SharingTypeAdded)



from py42.clients._archiveaccess import ArchiveContentStreamer
from py42.clients._archiveaccess import ArchiveExplorer


_FILE_SIZE_CALC_TIMEOUT = 10
_DEFAULT_BACKUP_SET_ID = u"1"


class ArchiveClient(object):
    """A module for getting information about backup archives on storage nodes along with
    functionality for streaming a file from backup.
    """

    def __init__(self, archive_accessor_factory, archive_service):
        self._archive_accessor_factory = archive_accessor_factory
        self._archive_service = archive_service

    def get_by_archive_guid(self, archive_guid):
        """Gets single archive information by GUID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Archive-get>`__

        Args:
            archive_guid (str): The GUID for the archive.

        Returns:
            :class:`py42.response.Py42Response`: A response containing archive
            information.
        """
        return self._archive_service.get_single_archive(archive_guid)

    def get_all_by_device_guid(self, device_guid):
        """Gets archive information for a device.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Archive-get>`__

        Args:
            device_guid (str): The GUID for the device.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response`
            objects that each contain a page of archives.
        """
        return self._archive_service.get_all_archives_from_value(
            device_guid, u"backupSourceGuid"
        )

    def stream_from_backup(
        self,
        file_paths,
        device_guid,
        destination_guid=None,
        archive_password=None,
        encryption_key=None,
        show_deleted=None,
        file_size_calc_timeout=_FILE_SIZE_CALC_TIMEOUT,
    ):
        """Streams a file from a backup archive to memory. This method uses the same endpoint
        as restoring from Console and therefore has all the same considerations.

        `Support Documentation <https://support.code42.com/Administrator/6/Monitoring_and_managing/Restore_files_from_the_Code42_console#Troubleshooting>`__

        Args:
            file_paths (str or list of str): The path or list of paths to the files or directories in
                the archive.
            device_guid (str): The GUID of the device the file belongs to.
            destination_guid (str, optional): The GUID of the destination that stores the backup
                of the file. If None, it will use the first destination GUID it finds for your
                device. 'destination_guid' may be useful if the file is missing from one of your
                destinations or if you want to optimize performance. Defaults to None.
            archive_password (str or None, optional): The password for the archive, if password-
                protected. This is only relevant to users with archive key password security. Defaults
                to None.
            encryption_key (str or None, optional): A custom encryption key for decrypting an archive's
                file contents, necessary for restoring files. This is only relevant to users with custom
                key archive security. Defaults to None.
            show_deleted (bool, optional): Set to True to include deleted files when restoring a directory.
                Defaults to None.
            file_size_calc_timeout (int, optional): Set to limit the amount of seconds spent calculating
                file sizes when crafting the request. Set to 0 or None to ignore file sizes altogether.
                Defaults to 10.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the streamed content.

        Usage example::

            stream_response = sdk.archive.stream_from_backup("/full/path/to/file.txt", "1234567890")
            with open("/path/to/save/file/to", "wb") as f:
                for chunk in stream_response.iter_content(chunk_size=128):
                    if chunk:
                        f.write(chunk)

        In certain cases, you will have to unzip the results::

            import zipfile
            with zipfile.ZipFile("downloaded_directory.zip", "r") as zf:
                zf.extractall(".")
        """
        archive_accessor = self._archive_accessor_factory.create_archive_accessor(
            device_guid,
            ArchiveContentStreamer,
            destination_guid=destination_guid,
            private_password=archive_password,
            encryption_key=encryption_key,
        )
        backup_set_id = self._select_backup_set_id(
            device_guid, archive_accessor.destination_guid
        )
        return archive_accessor.stream_from_backup(
            backup_set_id,
            file_paths,
            file_size_calc_timeout=file_size_calc_timeout,
            show_deleted=show_deleted,
        )

    def stream_to_device(
        self,
        file_paths,
        device_guid,
        accepting_device_guid,
        restore_path,
        destination_guid=None,
        archive_password=None,
        encryption_key=None,
        show_deleted=None,
        overwrite_existing_files=False,
        file_size_calc_timeout=_FILE_SIZE_CALC_TIMEOUT,
    ):
        """Streams a file from a backup archive to a specified device.

        Args:
            file_paths (str or list of str): The path or list of paths to the files or directories in
                the archive.
            device_guid (str): The GUID of the device the file belongs to.
            accepting_device_guid (str): The GUID of the device accepting the restore.
            restore_path (str, optional): The path on the accepting device where the restore will be
                saved. Alternatively, pass in the value ``ORIGINAL_LOCATION`` to restore the file
                to the original location, which may be the case if you are replacing a device.
            destination_guid (str, optional): The GUID of the destination that stores the backup
                of the file. If None, it will use the first destination GUID it finds for your
                device. 'destination_guid' may be useful if the file is missing from one of your
                destinations or if you want to optimize performance. Defaults to None.
            archive_password (str or None, optional): The password for the archive, if password-
                protected. This is only relevant to users with archive key password security. Defaults
                to None.
            encryption_key (str or None, optional): A custom encryption key for decrypting an archive's
                file contents, necessary for restoring files. This is only relevant to users with custom
                key archive security. Defaults to None.
            show_deleted (bool, optional): Set to True to include deleted files when restoring a directory.
                Defaults to None.
            overwrite_existing_files (bool, optional): to overwrite any existing files with the restored
                data. If False (the default), any existing files that match a path being restored will
                first get renamed.
            file_size_calc_timeout (int, optional): Set to limit the amount of seconds spent calculating
                file sizes when crafting the request. Set to 0 or None to ignore file sizes altogether.
                Defaults to 10.

        Returns:
            :class:`py42.response.Py42Response`.
        """
        explorer = self._archive_accessor_factory.create_archive_accessor(
            device_guid,
            ArchiveExplorer,
            destination_guid=destination_guid,
            private_password=archive_password,
            encryption_key=encryption_key,
        )
        backup_set_id = self._select_backup_set_id(
            device_guid, explorer.destination_guid
        )
        file_selections = explorer.create_file_selections(
            backup_set_id, file_paths, file_size_calc_timeout
        )
        pusher = self._archive_accessor_factory.create_archive_content_pusher(
            device_guid,
            accepting_device_guid,
            private_password=archive_password,
            encryption_key=encryption_key,
            destination_guid=explorer.destination_guid,
        )
        return pusher.stream_to_device(
            restore_path,
            accepting_device_guid,
            file_selections,
            backup_set_id,
            show_deleted,
            overwrite_existing_files,
        )

    def _select_backup_set_id(self, device_guid, destination_guid):
        backup_sets = self.get_backup_sets(device_guid, destination_guid)[u"backupSets"]
        if not backup_sets:
            return None
        for backup_set in backup_sets:
            backup_set_id = backup_set[u"backupSetId"]
            if str(backup_set_id) == _DEFAULT_BACKUP_SET_ID:
                return backup_set_id

        return backup_sets[0][u"backupSetId"]

    def get_backup_sets(self, device_guid, destination_guid):
        """Gets all backup set names/identifiers referring to a single destination for a specific
        device.
        `Learn more about backup sets. <https://support.code42.com/Administrator/Cloud/Configuring/Backup_sets>`__

        Args:
            device_guid (str): The GUID of the device to get backup sets for.
            destination_guid (str): The GUID of the destination containing the archive to get
                backup sets for.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the backup sets.
        """
        return self._archive_service.get_backup_sets(device_guid, destination_guid)

    def get_all_org_restore_history(self, days, org_id):
        """Gets all restore jobs from the past given days for the organization with the given ID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#RestoreHistory-get>`__

        Args:
            days (int): Number of days of restore history to retrieve.
            org_id (int): The identification number of the organization to get restore history for.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of restore history.
        """
        return self._archive_service.get_all_restore_history(days, u"orgId", org_id)

    def get_all_user_restore_history(self, days, user_id):
        """Gets all restore jobs from the past given days for the user with the given ID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#RestoreHistory-get>`__

        Args:
            days (int): Number of days of restore history to retrieve.
            user_id (int): The identification number of the user to get restore history for.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of restore history.
        """
        return self._archive_service.get_all_restore_history(days, u"userId", user_id)

    def get_all_device_restore_history(self, days, device_id):
        """Gets all restore jobs from the past given days for the device with the given ID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#RestoreHistory-get>`__

        Args:
            days (int): Number of days of restore history to retrieve.
            device_id (int): The identification number of the device to get restore history for.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of restore history.
        """
        return self._archive_service.get_all_restore_history(
            days, u"computerId", device_id
        )

    def update_cold_storage_purge_date(self, archive_guid, purge_date):
        """Updates the cold storage purge date for a specified archive.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#ColdStorage-put>`__

        Args:
            archive_guid (str): The identification number of the archive that should be updated
            purge_date (str): The date on which the archive should be purged in yyyy-MM-dd format

        Returns:
            :class:`py42.response.Py42Response`: the response from the ColdStorage API.
        """
        return self._archive_service.update_cold_storage_purge_date(
            archive_guid, purge_date
        )

    def get_all_org_cold_storage_archives(
        self,
        org_id,
        include_child_orgs=True,
        sort_key=u"archiveHoldExpireDate",
        sort_dir=u"asc",
    ):
        """Returns a detailed list of cold storage archive information for a given org ID.

        Args:
            org_id (str): The ID of a Code42 organization.
            include_child_orgs (bool, optional): Determines whether cold storage information from
             the Org's children is also returned. Defaults to True.
            sort_key (str, optional): Sets the property by which the returned results will be sorted.
             Choose from archiveHoldExpireDate, orgName, mountPointName, archiveBytes, and archiveType. Defaults to archiveHoldExpireDate.
            sort_dir (str, optional): Sets the order by which sort_key should be sorted. Choose from
             asc or desc. Defaults to asc.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of cold storage archive information.
        """
        return self._archive_service.get_all_org_cold_storage_archives(
            org_id=org_id,
            include_child_orgs=include_child_orgs,
            sort_key=sort_key,
            sort_dir=sort_dir,
        )



from collections import namedtuple

Clients = namedtuple(
    u"Clients",
    [
        u"alerts",
        u"archive",
        u"authority",
        u"detectionlists",
        u"securitydata",
        u"auditlogs",
        u"cases",
    ],
)



from py42.services import BaseService


class KeyValueStoreService(BaseService):
    def __init__(self, connection):
        super(KeyValueStoreService, self).__init__(connection)

    def get_stored_value(self, key):
        uri = u"/v1/{}".format(key)
        # this request doesn't expect json, unlike most.
        headers = {u"Accept": u"*/*"}
        return self._connection.get(uri, headers=headers)



from py42.sdk.queries.fileevents.file_event_query import FileEventFilterStringField
from py42.util import get_attribute_keys_from_class


class ExposureType(FileEventFilterStringField):
    """Class that filters events based on exposure type.

    Available options are provided as class attributes:
        - :attr:`ExposureType.SHARED_VIA_LINK`
        - :attr:`ExposureType.SHARED_TO_DOMAIN`
        - :attr:`ExposureType.APPLICATION_READ`
        - :attr:`ExposureType.CLOUD_STORAGE`
        - :attr:`ExposureType.REMOVABLE_MEDIA`
        - :attr:`ExposureType.IS_PUBLIC`
    """

    _term = u"exposure"

    SHARED_VIA_LINK = u"SharedViaLink"
    SHARED_TO_DOMAIN = u"SharedToDomain"
    APPLICATION_READ = u"ApplicationRead"
    CLOUD_STORAGE = u"CloudStorage"
    REMOVABLE_MEDIA = u"RemovableMedia"
    IS_PUBLIC = u"IsPublic"
    OUTSIDE_TRUSTED_DOMAINS = u"OutsideTrustedDomains"

    @staticmethod
    def choices():
        return get_attribute_keys_from_class(ExposureType)


class ProcessName(FileEventFilterStringField):
    """Class that filters events based on the process name involved in the exposure (applies to
    ``read by browser or other app`` events only).
    """

    _term = u"processName"


class ProcessOwner(FileEventFilterStringField):
    """Class that filters events based on the process owner that was involved in the exposure
    (applies to ``read by browser or other app`` events only).
    """

    _term = u"processOwner"


class RemovableMediaName(FileEventFilterStringField):
    """Class that filters events based on the name of the removable media involved in the exposure
    (applies to ``removable media`` events only).
    """

    _term = u"removableMediaName"


class RemovableMediaVendor(FileEventFilterStringField):
    """Class that filters events based on the vendor of the removable media device involved in the
    exposure (applies to ``removable media`` events only).
    """

    _term = u"removableMediaVendor"


class RemovableMediaMediaName(FileEventFilterStringField):
    """Class that filters events based on the name of the removable media (as reported by the
    vendor/device, usually very similar to RemovableMediaName) involved in the exposure (applies to
    ``removable media`` events only).
    """

    _term = u"removableMediaMediaName"


class RemovableMediaVolumeName(FileEventFilterStringField):
    """Class that filters events based on the name of the formatted volume (as reported by the
    operating system) of the removable media device involved in the exposure (applies to
    ``removable media`` events only).
    """

    _term = u"removableMediaVolumeName"


class RemovableMediaPartitionID(FileEventFilterStringField):
    """Class that filters events based on the unique identifier assigned (by the operating system)
    to the removable media involved in the exposure (applies to ``removable media`` events only).
    """

    _term = u"removableMediaPartitionId"


class RemovableMediaSerialNumber(FileEventFilterStringField):
    """Class that filters events based on the serial number of the connected hardware as reported
    by the operating system (applies to ``removable media`` events only).
    """

    _term = u"removableMediaSerialNumber"


class SyncDestination(FileEventFilterStringField):
    """Class that filters events based on the name of the cloud service the file is synced with
    (applies to ``synced to cloud service`` events only).

    Available options are provided as class attributes:
        - :attr:`SyncDestination.ICLOUD`
        - :attr:`SyncDestination.BOX`
        - :attr:`SyncDestination.BOX_DRIVE`
        - :attr:`SyncDestination.GOOGLE_DRIVE`
        - :attr:`SyncDestination.GOOGLE_BACKUP_AND_SYNC`
        - :attr:`SyncDestination.DROPBOX`
        - :attr:`SyncDestination.ONEDRIVE`
    """

    _term = u"syncDestination"

    ICLOUD = u"ICloud"
    BOX = u"Box"
    BOX_DRIVE = u"BoxDrive"
    GOOGLE_DRIVE = u"GoogleDrive"
    GOOGLE_BACKUP_AND_SYNC = u"GoogleBackupAndSync"
    DROPBOX = u"Dropbox"
    ONEDRIVE = u"OneDrive"

    @staticmethod
    def choices():
        return get_attribute_keys_from_class(SyncDestination)


class TabURL(FileEventFilterStringField):
    """Class that filters events based on all the URLs of the browser tabs at the time the file
    contents were read by the browser (applies to ``read by browser or other app`` events only).
    """

    _term = u"tabUrls"


class WindowTitle(FileEventFilterStringField):
    """Class that filters events based on the name of all the browser tabs or application windows that were
    open when a browser or other app event occurred (applies to ``read by browser or other app``
    events only).
    """

    _term = u"tabTitles"



from py42.util import get_attribute_keys_from_class


class RiskTags(object):
    """The available RiskTags for :meth:`~py42.clients.detectionlists.DetectionListsClient.add_user_risk_tags()`
    and :meth:`~py42.clients.detectionlists.DetectionListsClient.remove_user_risk_tags()`"""

    FLIGHT_RISK = u"FLIGHT_RISK"
    HIGH_IMPACT_EMPLOYEE = u"HIGH_IMPACT_EMPLOYEE"
    ELEVATED_ACCESS_PRIVILEGES = u"ELEVATED_ACCESS_PRIVILEGES"
    PERFORMANCE_CONCERNS = u"PERFORMANCE_CONCERNS"
    SUSPICIOUS_SYSTEM_ACTIVITY = u"SUSPICIOUS_SYSTEM_ACTIVITY"
    POOR_SECURITY_PRACTICES = u"POOR_SECURITY_PRACTICES"
    CONTRACT_EMPLOYEE = u"CONTRACT_EMPLOYEE"

    @staticmethod
    def choices():
        return get_attribute_keys_from_class(RiskTags)


class DetectionListsClient(object):
    """`Rest documentation <https://developer.code42.com/api/#tag/Detection-Lists>`__"""

    def __init__(
        self,
        user_profile_service,
        departing_employee_service,
        high_risk_employee_service,
    ):
        self._user_profile_service = user_profile_service
        self._departing_employee_service = departing_employee_service
        self._high_risk_employee_service = high_risk_employee_service

    @property
    def departing_employee(self):
        return self._departing_employee_service

    @property
    def high_risk_employee(self):
        return self._high_risk_employee_service

    def create_user(self, username):
        """Create a detection list profile for a user.
        `Rest Documentation <https://developer.code42.com/api/#operation/UserControllerV2_Create>`__

        Args:
            username (str): The Code42 username of the user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._user_profile_service.create(username)

    def get_user(self, username):
        """Get user details by username.
        `Rest Documentation <https://developer.code42.com/api/#operation/UserControllerV2_GetByUsername>`__

        Args:
            username (str): The Code42 username of the user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._user_profile_service.get(username)

    def get_user_by_id(self, user_id):
        """Get user details by user_id.
        `Rest Documentation <https://developer.code42.com/api/#operation/UserControllerV2_GetByUserId>`__

        Args:
            user_id (str or int): The Code42 userId of the user.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._user_profile_service.get_by_id(user_id)

    def update_user_notes(self, user_id, notes):
        """Add or update notes related to the user.
        `Rest Documentation <https://developer.code42.com/api/#operation/UserControllerV2_UpdateNotes>`__

        Args:
            user_id (str or int): The userUid of the user whose notes you want to update.
            notes (str): User profile notes.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._user_profile_service.update_notes(user_id, notes)

    def add_user_risk_tags(self, user_id, tags):
        """Add one or more risk factor tags.
        `Rest Documentation <https://developer.code42.com/api/#operation/UserControllerV2_AddRiskFactors>`__

        Args:
            user_id (str or int): The userUid of the user whose risk factor tag(s) you want to update.
            tags (str or list of str ): A single tag or multiple tags in a list to be added. For
                example: ``"tag1"`` or ``["tag1", "tag2"]``. For python version 2.X, pass ``u"str"``
                instead of ``"str"``.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._user_profile_service.add_risk_tags(user_id, tags)

    def remove_user_risk_tags(self, user_id, tags):
        """Remove one or more risk factor tags.
        `Rest Documentation <https://developer.code42.com/api/#operation/UserControllerV2_RemoveRiskFactors>`__

        Args:
            user_id (str or int): The userUid of the user whose risk factor tag(s) needs you want to remove.
            tags (str or list of str ): A single tag or multiple tags in a list to be removed. For
                example: ``"tag1"`` or ``["tag1", "tag2"]``. For python version 2.X, pass ``u"str"``
                instead of ``"str"``.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._user_profile_service.remove_risk_tags(user_id, tags)

    def add_user_cloud_alias(self, user_id, alias):
        """Add a cloud alias to a user.
        `Rest Documentation <https://developer.code42.com/api/#operation/UserControllerV2_AddCloudUsernames>`__

        Args:
            user_id (str or int): The userUid of the user whose alias you want to update.
            alias (str): The alias to be added.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._user_profile_service.add_cloud_alias(user_id, alias)

    def remove_user_cloud_alias(self, user_id, alias):
        """Remove a cloud alias from a user.
        `Rest Documentation <https://developer.code42.com/api/#operation/UserControllerV2_RemoveCloudUsernames>`__

        Args:
            user_id (str or int): The userUid of the user whose alias needs to be removed.
            alias (str): The alias to be removed.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._user_profile_service.remove_cloud_alias(user_id, alias)

    def refresh_user_scim_attributes(self, user_id):
        """Refresh SCIM attributes of a user.
        `Rest Documentation <https://developer.code42.com/api/#operation/UserControllerV2_RefreshUser>`__

        Args:
            user_id (str or int): The userUid of the user whose attributes you wish to refresh.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._user_profile_service.refresh(user_id)



from collections import namedtuple

from py42 import settings
from py42.clients.settings.org_settings import OrgSettings
from py42.exceptions import Py42Error
from py42.services import BaseService
from py42.services.util import get_all_pages

OrgSettingsResponse = namedtuple(
    u"OrgSettingsResponse", [u"error", u"org_response", u"org_settings_response"]
)


class OrgService(BaseService):
    """A service for interacting with Code42 organization APIs.

    Use the OrgService to create and retrieve organizations. You can also use it to block and
    deactivate organizations.
    """

    def create_org(self, org_name, org_ext_ref=None, notes=None, parent_org_uid=None):
        """Creates a new organization.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Org-post>`__

        Args:
            org_name (str): The name of the new organization.
            org_ext_ref (str, optional): External reference information,
                such as a serial number, asset tag, employee ID, or help desk issue ID. Defaults to
                None.
            notes (str, optional): Descriptive information about the organization. Defaults to None.
            parent_org_uid (int, optional): The org UID for the parent organization. Defaults to
                None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/Org/"
        data = {
            u"orgName": org_name,
            u"orgExtRef": org_ext_ref,
            u"notes": notes,
            u"parentOrgUid": parent_org_uid,
        }
        return self._connection.post(uri, json=data)

    def get_by_id(self, org_id, **kwargs):
        """Gets the organization with the given ID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Org-get>`__

        Args:
            org_id (int): An ID for an organization.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the organization.
        """
        uri = u"/api/Org/{}".format(org_id)
        return self._connection.get(uri, params=kwargs)

    def get_by_uid(self, org_uid, **kwargs):
        """Gets the organization with the given UID.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Org-get>`__

        Args:
            org_uid (str): A UID for an organization.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the organization.
        """
        uri = u"/api/Org/{}".format(org_uid)
        params = dict(idType=u"orgUid", **kwargs)
        return self._connection.get(uri, params=params)

    def get_page(self, page_num, page_size=None, **kwargs):
        """Gets an individual page of organizations.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Org-get>`__

        Args:
            page_num (int): The page number to request.
            page_size (int, optional): The number of organizations to return per page.
                Defaults to `py42.settings.items_per_page`.
            kwargs (dict, optional): Additional advanced-user arguments. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        page_size = page_size or settings.items_per_page
        uri = u"/api/Org"
        params = dict(pgNum=page_num, pgSize=page_size, **kwargs)
        return self._connection.get(uri, params=params)

    def get_all(self, **kwargs):
        """Gets all organizations.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Org-get>`__

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of organizations.
        """
        return get_all_pages(self.get_page, u"orgs", **kwargs)

    def block(self, org_id):
        """Blocks the organization with the given org ID as well as its child organizations. A
        blocked organization will not allow any of its users or devices to log in. New
        registrations will be rejected and all currently logged in clients will be logged out.
        Backups continue for any devices that are still active.
        `Rest Documentation <https://console.us.code42.com/apidocviewer/#OrgBlock-put>`__

        Args:
            org_id (int): An ID for an organization.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/OrgBlock/{}".format(org_id)
        return self._connection.put(uri)

    def unblock(self, org_id):
        """Removes a block, if one exists, on an organization and its descendants with the given
        ID. All users in the organization remain blocked until they are unblocked individually.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#OrgBlock-delete>`__

        Args:
            org_id (int): An ID for an organization.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/OrgBlock/{}".format(org_id)
        return self._connection.delete(uri)

    def deactivate(self, org_id):
        """Deactivates the organization with the given ID, including all users, plans, and
        devices. Backups stop and archives move to cold storage.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#OrgDeactivation-put>`__

        Args:
            org_id (int): An ID for an organization.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/OrgDeactivation/{}".format(org_id)
        return self._connection.put(uri)

    def reactivate(self, org_id):
        """Reactivates the organization with the given ID. Backups are *not* restarted
        automatically.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#OrgDeactivation-delete>`__

        Args:
            org_id (int): An ID for an organization.

        Returns:
            :class:`py42.response.Py42Response`
        """
        uri = u"/api/OrgDeactivation/{}".format(org_id)
        return self._connection.delete(uri)

    def get_current(self, **kwargs):
        """Gets the organization for the currently signed-in user.
        `REST Documentation <https://console.us.code42.com/apidocviewer/#Org-get>`__

        Returns:
            :class:`py42.response.Py42Response`: A response containing the organization for the
            currently signed-in user.
        """
        uri = u"/api/Org/my"
        return self._connection.get(uri, params=kwargs)

    def get_agent_state(self, org_id, property_name):
        """Gets the agent state of the devices in the org.
            `REST Documentation <https://console.us.code42.com/swagger/index.html?urls.primaryName=v14#/agent-state/AgentState_ViewByDeviceGuid>`__

            Args:
                org_id (str): The org's identifier.
                property_name (str): The name of the property to retrieve (e.g. `fullDiskAccess`).

            Returns:
                :class:`py42.response.Py42Response`: A response containing settings information.
            """
        uri = u"/api/v14/agent-state/view-by-organization-id"
        params = {u"orgId": org_id, u"propertyName": property_name}
        return self._connection.get(uri, params=params)

    def get_agent_full_disk_access_states(self, org_id):
        """Gets the full disk access status for devices in an org.
            `REST Documentation <https://console.us.code42.com/swagger/index.html?urls.primaryName=v14#/agent-state/AgentState_ViewByDeviceGuid>`__

            Args:
                org_id (str): The org's identifier.

            Returns:
                :class:`py42.response.Py42Response`: A response containing settings information.
            """
        return self.get_agent_state(org_id, u"fullDiskAccess")

    def get_settings(self, org_id):
        """Gets setting data for an org and returns an `OrgSettingsManager` for the target org.

        Args:
            org_id (int,str): The identifier of the org.

        Returns:
            :class:`py42.clients._settings_managers.OrgSettings`: A class to help manage org settings.
                """
        org_settings = self.get_by_id(
            org_id, incSettings=True, incDeviceDefaults=True, incInheritedOrgInfo=True
        )
        uri = u"/api/OrgSetting/{}".format(org_id)
        t_settings = self._connection.get(uri)
        return OrgSettings(org_settings.data, t_settings.data)

    def update_settings(self, org_settings):
        """Updates an org's settings based on changes to the passed in `OrgSettings` instance.

        Args:
            org_settings (`OrgSettings`): An `OrgSettings` instance with desired modifications to settings.

        Returns:
            :class:`py42.services.orgs.OrgSettings`: A namedtuple containing the result of the setting change api calls.
        """
        org_id = org_settings.org_id
        error = False
        org_settings_response = org_response = None

        if org_settings.packets:
            uri = u"/api/OrgSetting/{}".format(org_id)
            payload = {"packets": org_settings.packets}
            try:
                org_settings_response = self._connection.put(uri, json=payload)
            except Py42Error as ex:
                error = True
                org_settings_response = ex

        if org_settings.changes:
            uri = "/api/Org/{}".format(org_id)
            try:
                org_response = self._connection.put(uri, json=org_settings.data)
            except Py42Error as ex:
                error = True
                org_response = ex
        return OrgSettingsResponse(
            error=error,
            org_response=org_response,
            org_settings_response=org_settings_response,
        )



from py42._compat import str
from py42.sdk.queries import BaseQuery
from py42.sdk.queries.query_filter import create_filter_group
from py42.sdk.queries.query_filter import create_query_filter
from py42.sdk.queries.query_filter import create_within_the_last_filter_group
from py42.sdk.queries.query_filter import QueryFilterStringField
from py42.sdk.queries.query_filter import QueryFilterTimestampField


class FileEventQuery(BaseQuery):
    """Helper class for building Code42 Forensic Search queries.

    A FileEventQuery instance's ``all()`` and ``any()`` take one or more
    :class:`~py42.sdk.queries.query_filter.FilterGroup` objects to construct a query that
    can be passed to the :meth:`FileEventService.search()` method. ``all()`` returns results
    that match all of the provided filter criteria, ``any()`` will return results that
    match any of the filters.

    For convenience, the :class:`FileEventQuery` constructor does the same as ``all()``.

    Usage example::

        email_filter = EmailSender.is_in(["test.user@example.com", "test.sender@example.com"])
        exposure_filter = ExposureType.exists()
        query = FileEventQuery.all(email_filter, exposure_filter)

    """

    def __init__(self, *args, **kwargs):
        super(FileEventQuery, self).__init__(*args, **kwargs)
        self.sort_key = u"eventId"

    def __str__(self):
        groups_string = u",".join(
            str(group_item) for group_item in self._filter_group_list
        )
        if self.page_token is not None:
            paging_prop = u'"pgToken":"{}"'.format(self.page_token)

        else:
            paging_prop = u'"srtDir":"{}", "srtKey":"{}", "pgNum":{}'.format(
                self.sort_direction, self.sort_key, self.page_number,
            )
        json = u'{{"groupClause":"{0}", "groups":[{1}], {2}, "pgSize":{3}}}'.format(
            self._group_clause, groups_string, paging_prop, self.page_size
        )
        return json

    def __iter__(self):
        filter_group_list = [dict(item) for item in self._filter_group_list]
        output_dict = {
            u"groupClause": self._group_clause,
            u"groups": filter_group_list,
            u"pgNum": self.page_number,
            u"pgSize": self.page_size,
            u"srtDir": self.sort_direction,
            u"srtKey": self.sort_key,
            u"pgToken": self.page_token,
        }
        for key in output_dict:
            yield key, output_dict[key]


def create_exists_filter_group(term):
    """Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` to find events where
    filter data exists. Useful for creating ``EXISTS`` filters that are not yet supported
    in py42 or programmatically crafting filter groups.

    Args:
        term (str): The term of the filter.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """
    filter_list = [create_query_filter(term, u"EXISTS")]
    return create_filter_group(filter_list, u"AND")


def create_not_exists_filter_group(term):
    """Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` to find events where
    filter data does not exist. Useful for creating ``DOES_NOT_EXIST`` filters that are
    not yet supported in py42 or programmatically crafting filter groups.

    Args:
        term (str): The term of the filter.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """
    filter_list = [create_query_filter(term, u"DOES_NOT_EXIST")]
    return create_filter_group(filter_list, u"AND")


def create_greater_than_filter_group(term, value):
    """Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for matching file
    events where the value with key ``term`` is greater than the given value. Useful for
    creating ``GREATER_THAN`` filters that are not yet supported in py42 or programmatically
    crafting filter groups.

    Args:
        term (str): The term of the filter.
        value (str or int): The value used to filter file events.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """
    filter_list = [create_query_filter(term, u"GREATER_THAN", value)]
    return create_filter_group(filter_list, u"AND")


def create_less_than_filter_group(term, value):
    """Creates a :class:`~py42.sdk.queries.query_filter.FilterGroup` for matching file
    events where the value with key ``term`` is less than the given value. Useful for creating
    ``LESS_THAN`` filters that are not yet supported in py42 or programmatically crafting
    filter groups.

    Args:
        term (str): The term of the filter.
        value (str or int): The value used to filter file events.

    Returns:
        :class:`~py42.sdk.queries.query_filter.FilterGroup`
    """
    filter_list = [create_query_filter(term, u"LESS_THAN", value)]
    return create_filter_group(filter_list, u"AND")


class FileEventFilterStringField(QueryFilterStringField):
    """Helper class for creating filters with the ``EXISTS``/``NOT_EXISTS`` filter clauses."""

    @classmethod
    def exists(cls):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` to find events
        where filter data exists.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        return create_exists_filter_group(cls._term)

    @classmethod
    def not_exists(cls):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` to find events
        where filter data does not exist.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        return create_not_exists_filter_group(cls._term)


class FileEventFilterComparableField(object):
    """Helper class for creating filters with the ``GREATER_THAN``/``LESS_THAN`` filter clauses."""

    _term = u"override_boolean_field_name"

    @classmethod
    def greater_than(cls, value):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` to find events
        where filter data is greater than the provided value.

        Args:
            value (str or int or float): The value used to filter file events.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        value = int(value)
        return create_greater_than_filter_group(cls._term, value)

    @classmethod
    def less_than(cls, value):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` to find events
        where filter data is less than than the provided value.

        Args:
            value (str or int or float): The value used to filter file events.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        value = int(value)
        return create_less_than_filter_group(cls._term, value)


class FileEventFilterTimestampField(QueryFilterTimestampField):
    @classmethod
    def within_the_last(cls, value):
        """Returns a :class:`~py42.sdk.queries.query_filter.FilterGroup` that is useful
        for finding results where the key ``self._term`` is a timestamp-related term,
        such as ``EventTimestamp._term``, and ``value`` is one of it's accepted values,
        such as one of the values in ``EventTimestamp.choices()``.

        Args:
            value (str): The value used to filter file events.

        Returns:
            :class:`~py42.sdk.queries.query_filter.FilterGroup`
        """
        return create_within_the_last_filter_group(cls._term, value)



from py42._compat import string_type


def set_val(d, keys, value):
    """Helper for setting nested values from a dict based on a list of keys."""
    d = get_val(d, keys[:-1])
    d[keys[-1]] = value


def get_val(d, keys):
    """Helper for getting nested values from a dict based on a list of keys."""
    for key in keys:
        d = d[key]
    return d


def show_change(val1, val2):
    if isinstance(val1, string_type):
        val1 = u'"{}"'.format(val1)
    if isinstance(val2, string_type):
        val2 = u'"{}"'.format(val2)
    return u"{} -> {}".format(val1, val2)


class BaseSettingProperty(object):
    def __init__(self, name, location):
        self.name = name
        self.location = location
        self.init_val = None

    def _register_change(self, instance, orig_val, new_val):
        name = self.name.lstrip(u"_")
        if self.init_val is None:
            self.init_val = orig_val
        if self.init_val == new_val:
            if name in instance.changes:
                instance.changes.pop(name)
        else:
            instance.changes[name] = show_change(self.init_val, new_val)


class SettingProperty(BaseSettingProperty):
    """Descriptor class to help manage changes to nested dict values. Assumes attributes
    being managed are on a UserDict/UserList subclass.

    Args:
        name (str): name of attribute this class manages (changes will be registered with this name).
        location (list): list of keys defining the location of the value being managed in the managed class.
        get_converter (func, optional): function to convert retrieved values to preferred format. Defaults to no conversion.
        set_converter (func, optional): function to convert values being set to preferred format. Defaults to no conversion.
    """

    def __init__(
        self,
        name,
        location,
        get_converter=None,
        set_converter=None,
        inheritance_attr=None,
    ):
        super(SettingProperty, self).__init__(name, location)
        self.get_converter = get_converter
        self.set_converter = set_converter
        self.inheritance_attr = inheritance_attr

    def __get__(self, instance, owner):
        val = get_val(instance.data, self.location)
        if isinstance(val, dict) and u"#text" in val.keys():
            val = val[u"#text"]
        return self.get_converter(val) if self.get_converter is not None else val

    def __set__(self, instance, new_val):
        converted_new_val = (
            self.set_converter(new_val) if self.set_converter is not None else new_val
        )
        orig_val = get_val(instance.data, self.location)

        # if locked, value is a dict with '#text' as the _real_ value key. Some properties will have dicts with @nil: true, these should stay dicts, as the whole dict is the property value.
        if isinstance(orig_val, dict) and u"#text" in orig_val.keys():
            location = self.location + [u"#text"]
            orig_val = orig_val[u"#text"]
        else:
            location = self.location

        self._register_change(instance, orig_val, converted_new_val)
        if self.inheritance_attr is not None and getattr(
            instance, self.inheritance_attr
        ):
            setattr(instance, self.inheritance_attr, False)
        set_val(instance.data, location, converted_new_val)


class TSettingProperty(object):
    """Descriptor class to help manage transforming t_setting packet values. Assumes t_setting
    dict is stored in `._t_settings` attribute on managed instances.

    Args:
        name (str): name of attribute this class manages (changes will be registered with this name).
        key (str): name of t_setting packet this class is managing.
    """

    def __init__(self, name, key, get_converter=None, set_converter=None):
        self.name = name
        self.key = key
        self.get_converter = get_converter
        self.set_converter = set_converter
        self.init_val = None

    def __get__(self, instance, owner):
        if self.key in instance._packets:
            packet = instance._packets[self.key]
        else:
            packet = instance._t_settings.get(self.key)
        if packet is None:
            return None
        return (
            self.get_converter(packet[u"value"])
            if self.get_converter is not None
            else packet[u"value"]
        )

    def __set__(self, instance, val):
        val = self.set_converter(val) if self.set_converter is not None else val
        packet = {u"key": self.key, u"value": val, u"locked": False}
        instance._packets[self.key] = packet
        self._register_change(instance, val)

    def _register_change(self, instance, val):
        if self.init_val is None:
            packet = instance._t_settings.get(self.key)
            if packet is None:
                self.init_val = None
            else:
                self.init_val = packet[u"value"]
        if self.init_val == val:
            if self.name in instance.changes:
                instance.changes.pop(self.name)
        else:
            instance.changes[self.name] = show_change(self.init_val, val)


def check_lock(locked_attr):
    """Decorator for instance methods that enforces a property to be read-only if the
    `.locked` attribute of the provided `locked_attr` on the instance returns True.
    """

    def f(method):
        def func(_self, *args):
            if getattr(_self, locked_attr).locked:
                raise AttributeError("Unable to modify locked backup set.")
            else:
                return method(_self, *args)

        return func

    return f



from py42.services import BaseService


class SecurityDataService(BaseService):
    def get_security_event_locations(self, user_uid):
        uri = u"/c42api/v3/SecurityEventsLocation"
        params = {u"userUid": user_uid}
        return self._connection.get(uri, params=params)



from __future__ import print_function

import json
from datetime import datetime

from py42._compat import str

_MICROSECOND_FORMAT = u"%Y-%m-%dT%H:%M:%S.%fZ"
DATE_STR_FORMAT = u"%Y-%m-%d %H:%M:%S"


def format_json(json_string):
    """Converts a minified JSON str to a prettified JSON str.

    Args:
        json_string (str): A str representing minified JSON.

    Returns:
        (str): A str representing prettified JSON.
    """
    parsed = json.loads(json_string)
    return json.dumps(parsed, indent=4)


def print_response(response, label=None):
    """Prints a :class:`py42.response.Py42Response` as prettified JSON. If unable to load, it
    prints the given response.

    Args:
        response (:class:`py42.response.Py42Response`): The response to print.
        label (str, optional): A label at the beginning of the printed text. Defaults to None.
    """
    if label:
        print(label, end=u" ")
    try:
        print(format_json(response.text))
    except ValueError:
        print(response)


def convert_timestamp_to_str(timestamp):
    """Converts the given POSIX timestamp to a date str. The format matches strftime
    directives %Y-%m-%dT%H:%M:%S.%f.

    Args:
        timestamp (float or int): A POSIX timestamp.

    Returns:
        (str): A str representing the given timestamp. Example output looks like
        '2020-03-25T15:29:04.465Z'.
    """
    date = datetime.utcfromtimestamp(timestamp)
    return convert_datetime_to_timestamp_str(date)


def convert_datetime_to_timestamp_str(date):
    """Converts the given datetime to a formatted date str. The format matches strftime
    directives %Y-%m-%dT%H:%M:%S.%f.

    Args:
        date (datetime): The datetime object to convert.

    Returns:
        (str): A str representing the given date. Example output looks like
        '2020-03-25T15:29:04.465Z'.
    """
    prefix = date.strftime(u"%Y-%m-%dT%H:%M:%S.%f")[:-3]
    return u"{}Z".format(prefix)


def convert_datetime_to_epoch(date):
    return (date - datetime.utcfromtimestamp(0)).total_seconds()


def format_dict(dict_, label=None):
    indented_dict = json.dumps(dict_, indent=4)
    if label:
        return u"{} {}".format(label, indented_dict)
    return indented_dict


def get_attribute_keys_from_class(cls):
    """Returns attribute names for the given class.

    Args:
        cls (class): The class to obtain attributes from.

    Returns:
        (list): A list containing the attribute names of the given class.
    """
    return [
        cls().__getattribute__(attr)
        for attr in dir(cls)
        if not callable(cls().__getattribute__(attr)) and not attr.startswith(u"_")
    ]


def to_list(value):
    if not value:
        return []
    if not isinstance(value, (list, tuple)):
        return [value]
    return value


def parse_timestamp_to_milliseconds_precision(timestamp):
    if isinstance(timestamp, (int, float)):
        return convert_timestamp_to_str(timestamp)
    elif isinstance(timestamp, datetime):
        return convert_datetime_to_timestamp_str(timestamp)
    elif isinstance(timestamp, str):
        return convert_datetime_to_timestamp_str(
            datetime.strptime(timestamp, DATE_STR_FORMAT)
        )


def parse_timestamp_to_microseconds_precision(timestamp):
    if isinstance(timestamp, (int, float)):
        return datetime.utcfromtimestamp(timestamp).strftime(_MICROSECOND_FORMAT)
    elif isinstance(timestamp, datetime):
        return timestamp.strftime(_MICROSECOND_FORMAT)
    elif isinstance(timestamp, str):
        return datetime.strptime(timestamp, DATE_STR_FORMAT).strftime(
            _MICROSECOND_FORMAT
        )



import json

from py42 import settings
from py42._compat import str
from py42.sdk.queries.query_filter import create_eq_filter_group
from py42.services import BaseService
from py42.services.util import get_all_pages


class AlertService(BaseService):
    _uri_prefix = u"/svc/api/v1/{0}"

    _CREATED_AT = u"CreatedAt"
    _RULE_METADATA = u"ruleMetadata"
    _SEARCH_KEY = u"alerts"

    def __init__(self, connection, user_context):
        super(AlertService, self).__init__(connection)
        self._user_context = user_context

    def search(self, query, page_num=1, page_size=None):
        query.page_number = page_num - 1
        if page_size:
            query.page_size = page_size
        query = self._add_tenant_id_if_missing(query)
        uri = self._uri_prefix.format(u"query-alerts")
        return self._connection.post(uri, data=query)

    def get_search_page(self, query, page_num, page_size):
        query.page_number = page_num - 1
        query.page_size = page_size
        uri = self._uri_prefix.format(u"query-alerts")
        query = self._add_tenant_id_if_missing(query)
        return self._connection.post(uri, data=query)

    def search_all_pages(self, query):
        return get_all_pages(
            self.get_search_page,
            self._SEARCH_KEY,
            query=query,
            page_size=query.page_size,
        )

    def get_details(self, alert_ids):
        if not isinstance(alert_ids, (list, tuple)):
            alert_ids = [alert_ids]
        tenant_id = self._user_context.get_current_tenant_id()
        uri = self._uri_prefix.format(u"query-details")
        data = {u"tenantId": tenant_id, u"alertIds": alert_ids}
        results = self._connection.post(uri, json=data)
        return _convert_observation_json_strings_to_objects(results)

    def update_state(self, state, alert_ids, note=""):
        if not isinstance(alert_ids, (list, tuple)):
            alert_ids = [alert_ids]
        note = note or ""
        tenant_id = self._user_context.get_current_tenant_id()
        uri = self._uri_prefix.format(u"update-state")
        data = {
            u"tenantId": tenant_id,
            u"alertIds": alert_ids,
            u"note": note,
            u"state": state,
        }
        return self._connection.post(uri, json=data)

    def _add_tenant_id_if_missing(self, query):
        query_dict = json.loads(str(query))
        tenant_id = query_dict.get(u"tenantId", None)
        if tenant_id is None:
            query_dict[u"tenantId"] = self._user_context.get_current_tenant_id()
            return json.dumps(query_dict)
        else:
            return str(query)

    def get_rules_page(
        self, page_num, groups=None, sort_key=None, sort_direction=None, page_size=None
    ):
        # This API expects the first page to start with zero.
        page_num = page_num - 1
        page_size = page_size or settings.items_per_page
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"groups": groups or [],
            u"groupClause": u"AND",
            u"pgNum": page_num,
            u"pgSize": page_size,
            u"srtKey": sort_key,
            u"srtDirection": sort_direction,
        }
        uri = self._uri_prefix.format(u"rules/query-rule-metadata")
        return self._connection.post(uri, json=data)

    def get_all_rules(self, sort_key=_CREATED_AT, sort_direction=u"DESC"):
        return get_all_pages(
            self.get_rules_page,
            self._RULE_METADATA,
            groups=None,
            sort_key=sort_key,
            sort_direction=sort_direction,
        )

    def get_all_rules_by_name(
        self, rule_name, sort_key=_CREATED_AT, sort_direction=u"DESC"
    ):
        return get_all_pages(
            self.get_rules_page,
            self._RULE_METADATA,
            groups=[json.loads(str(create_eq_filter_group(u"Name", rule_name)))],
            sort_key=sort_key,
            sort_direction=sort_direction,
        )

    def get_rule_by_observer_id(
        self, observer_id, sort_key=_CREATED_AT, sort_direction=u"DESC"
    ):
        results = get_all_pages(
            self.get_rules_page,
            self._RULE_METADATA,
            groups=[
                json.loads(str(create_eq_filter_group(u"ObserverRuleId", observer_id)))
            ],
            sort_key=sort_key,
            sort_direction=sort_direction,
        )
        return next(results)

    def update_note(self, alert_id, note):
        tenant_id = self._user_context.get_current_tenant_id()
        uri = self._uri_prefix.format(u"add-note")
        data = {
            u"tenantId": tenant_id,
            u"alertId": alert_id,
            u"note": note,
        }
        return self._connection.post(uri, json=data)

    def get_aggregate_data(self, alert_id):
        uri = self._uri_prefix.format(u"query-details-aggregate")
        data = {
            u"alertId": alert_id,
        }
        return self._connection.post(uri, json=data)


def _convert_observation_json_strings_to_objects(results):
    for alert in results[u"alerts"]:
        if u"observations" in alert:
            for observation in alert[u"observations"]:
                try:
                    observation[u"data"] = json.loads(observation[u"data"])
                except Exception:
                    continue
    return results



from datetime import datetime

from py42.util import get_attribute_keys_from_class
from py42.util import parse_timestamp_to_milliseconds_precision


class CaseStatus(object):
    OPEN = u"OPEN"
    CLOSED = u"CLOSED"

    @staticmethod
    def choices():
        return get_attribute_keys_from_class(CaseStatus)


class CasesClient(object):
    """A client to expose cases API.

    `Rest documentation <https://developer.code42.com/api/#tag/Cases>`__
    """

    def __init__(self, cases_service, cases_file_event_service):
        self._cases_service = cases_service
        self._file_events = cases_file_event_service

    @property
    def file_events(self):
        """A collection of methods for managing file events associated with a given case.

        Returns:
            :class:`py42.services.casesfileevents.CasesFileEventsService`
        """
        return self._file_events

    def create(
        self, name, subject=None, assignee=None, description=None, findings=None
    ):
        """Creates a new case.
        `Rest documentation <https://developer.code42.com/api/#operation/createCaseUsingPOST>`__

        Args:
            name (str): Name of the case.
            subject (str, optional): User UID of a subject of a case.
            assignee (str, optional): User UID of the assignee.
            description (str, optional): Description of the case
            findings (str, optional): Observations of the case.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._cases_service.create(
            name,
            subject=subject,
            assignee=assignee,
            description=description,
            findings=findings,
        )

    def get_page(
        self,
        page_num,
        name=None,
        status=None,
        min_create_time=None,
        max_create_time=None,
        min_update_time=None,
        max_update_time=None,
        subject=None,
        assignee=None,
        page_size=100,
        sort_direction=u"asc",
        sort_key=u"number",
        **kwargs
    ):
        """Gets individual page of cases.
        `Rest documentation <https://developer.code42.com/api/#operation/getCasesUsingGET>`__

        Args:
            page_num (int): The page number to request.
            name (str, optional): Filter results by case name, matches partial names. Defaults to None.
            status (str, optional): Filter results by case status. `CaseStatus.OPEN` or `CaseStatus.CLOSED` Defaults to None.
            min_create_time (str or int or float or datetime, optional): Filter results by case creation time, start time.
                 str format %Y-%m-%d %H:%M:%S. Defaults to None.
            max_create_time (str or int or float or datetime, optional): Filter results by case creation time, end time.
                 str format %Y-%m-%d %H:%M:%S. Defaults to None.
            min_update_time (str or int or float or datetime, optional): Filter results by last updated time, start time.
                 str format %Y-%m-%d %H:%M:%S. Defaults to None.
            max_update_time (str or int or float or datetime, optional): Filter results by last updated time, end time.
                 str format %Y-%m-%d %H:%M:%S. Defaults to None.
            subject (str, optional): Filter results based on User UID of a subject of a case. Defaults to None.
            assignee (str, optional): Filter results based on User UID of an assignee of a case. Defaults to None.
            page_size (int, optional): Number of results to return per page. Defaults to 100.
            sort_direction (str, optional): The direction on which to sort the response,
                based on the corresponding sort key. `asc` or `desc`. Defaults to `asc`.
            sort_key (str, optional): Values on which the response will be sorted. Defaults to "number".
                Available options are `name`, `number`, `createdAt`, `updatedAt`, `status`, `assigneeUsername`, `subjectUsername`.

        Returns:
            :class:`py42.response.Py42Response`
        """

        created_at = _make_range(min_create_time, max_create_time)
        updated_at = _make_range(min_update_time, max_update_time)

        return self._cases_service.get_page(
            page_num,
            name=name,
            status=status,
            created_at=created_at,
            updated_at=updated_at,
            subject=subject,
            assignee=assignee,
            page_size=page_size,
            sort_direction=sort_direction,
            sort_key=sort_key,
            **kwargs
        )

    def get_all(
        self,
        name=None,
        status=None,
        min_create_time=None,
        max_create_time=None,
        min_update_time=None,
        max_update_time=None,
        subject=None,
        assignee=None,
        page_size=100,
        sort_direction=u"asc",
        sort_key=u"number",
        **kwargs
    ):
        """Gets all cases.
        `Rest documentation <https://developer.code42.com/api/#operation/getCasesUsingGET>`__

        Args:
            name (str, optional): Filter results by case name, matches partial names. Defaults to None.
            status (str, optional): Filter results by case status. `CaseStatus.OPEN` or `CaseStatus.CLOSED` Defaults to None.
            min_create_time (str or int or float or datetime, optional): Filter results by case creation time, start time.
                 str format %Y-%m-%d %H:%M:%S. Defaults to None.
            max_create_time (str or int or float or datetime, optional): Filter results by case creation time, end time.
                 str format %Y-%m-%d %H:%M:%S. Defaults to None.
            min_update_time (str or int or float or datetime, optional): Filter results by last updated time, start time.
                 str format %Y-%m-%d %H:%M:%S. Defaults to None.
            max_update_time (str or int or float or datetime, optional): Filter results by last updated time, end time.
                 str format %Y-%m-%d %H:%M:%S. Defaults to None.
            subject (str, optional): Filter results based on User UID of a subject of a case. Defaults to None.
            assignee (str, optional): Filter results based on User UID of an assignee of a case. Defaults to None.
            page_size (int, optional): Number of results to return per page. Defaults to 100.
            sort_direction (str, optional): The direction on which to sort the response,
                based on the corresponding sort key. `asc` or `desc`. Defaults to `asc`.
            sort_key (str, optional): Values on which the response will be sorted. Defaults to "number".
                Available options are `name`, `number`, `createdAt`, `updatedAt`, `status`, `assigneeUsername`, `subjectUsername`.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of cases.
        """

        created_at = _make_range(min_create_time, max_create_time)
        updated_at = _make_range(min_update_time, max_update_time)

        return self._cases_service.get_all(
            name=name,
            status=status,
            created_at=created_at,
            updated_at=updated_at,
            subject=subject,
            assignee=assignee,
            page_size=page_size,
            sort_direction=sort_direction,
            sort_key=sort_key,
            **kwargs
        )

    def get(self, case_number):
        """Retrieve case details by case number.
        `Rest documentation <https://developer.code42.com/api/#operation/getCaseUsingGET>`__

        Args:
            case_number (int): Case number of the case.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._cases_service.get(case_number)

    def export_summary(self, case_number):
        """Provides case summary to download as a PDF file.
        `Rest documentation <https://developer.code42.com/api/#operation/pdfExportUsingGET>`__


        Args:
            case_number (int): Case number of the case.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._cases_service.export_summary(case_number)

    def update(
        self,
        case_number,
        name=None,
        subject=None,
        assignee=None,
        description=None,
        findings=None,
        status=None,
    ):
        """Updates case details for the given case number.
        `Rest documentation <https://developer.code42.com/api/#operation/updateCaseUsingPUT>`__


        Args:
            case_number (int): Case number of the case.
            name (str, optional): Name of the case. Defaults to None.
            subject (str, optional): A subject of the case. Defaults to None.
            assignee (str, optional): User UID of the assignee. Defaults to None.
            description (str, optional): Description of the case. Defaults to None.
            findings (str, optional): Notes on the case. Defaults to None.
            status (str, optional): Status of the case. `CaseStatus.OPEN` or `CaseStatus.CLOSED`.
                Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._cases_service.update(
            case_number,
            name=name,
            subject=subject,
            assignee=assignee,
            description=description,
            findings=findings,
            status=status,
        )


def _make_range(begin_time, end_time):
    if not begin_time and not end_time:
        return None
    if not begin_time:
        begin_time = datetime.utcfromtimestamp(0)
    if not end_time:
        end_time = datetime.utcnow()
    end = parse_timestamp_to_milliseconds_precision(end_time)
    start = parse_timestamp_to_milliseconds_precision(begin_time)
    return "{}/{}".format(start, end)



class AuditLogsClient(object):
    """`Rest documentation <https://developer.code42.com/api/#tag/Audit-Log>`__"""

    def __init__(self, audit_log_service):
        self._audit_log_service = audit_log_service

    def get_page(
        self,
        page_num=1,
        page_size=None,
        begin_time=None,
        end_time=None,
        event_types=None,
        user_ids=None,
        usernames=None,
        user_ip_addresses=None,
        affected_user_ids=None,
        affected_usernames=None,
        **kwargs
    ):
        """Retrieve a page of audit logs, filtered based on given arguments.

        Note: `page_num` here can be used same way as other methods that have a
        `page_num` parameter in py42. However, under the hood, it subtracts one from
        the given `page_num` in the implementation as the Code42 Audit-Logs API expects
        the start page to be zero.
        `Rest Documentation <https://developer.code42.com/api/#operation/rpc/search/search-audit-log>`__

        Args:
            page_num (int, optional): The page number to get. Defaults to 1.
            page_size (int, optional): The number of items per page. Defaults to `py42.settings.items_per_page`.
            begin_time (int or float or str or datetime, optional): Timestamp in milliseconds or
                str format "yyyy-MM-dd HH:MM:SS" or a datetime instance. Defaults to None.
            end_time (int or float or str or datetime, optional): Timestamp in milliseconds or
                str format "yyyy-MM-dd HH:MM:SS" or a datetime instance. Defaults to None.
            event_types (str or list, optional): A str or list of str of valid event types. Defaults to None.
            user_ids (str or list, optional): A str or list of str of Code42 userUids. Defaults to None.
            usernames (str or list, optional): A str or list of str of Code42 usernames. Defaults to None.
            user_ip_addresses (str or list, optional): A str or list of str of user ip addresses. Defaults to None.
            affected_user_ids (str or list, optional): A str or list of str of affected Code42 userUids. Defaults to None.
            affected_usernames (str or list, optional): A str or list of str of affected Code42 usernames. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._audit_log_service.get_page(
            page_num=page_num,
            page_size=page_size,
            begin_time=begin_time,
            end_time=end_time,
            event_types=event_types,
            user_ids=user_ids,
            usernames=usernames,
            user_ip_addresses=user_ip_addresses,
            affected_user_ids=affected_user_ids,
            affected_usernames=affected_usernames,
            **kwargs
        )

    def get_all(
        self,
        begin_time=None,
        end_time=None,
        event_types=None,
        user_ids=None,
        usernames=None,
        user_ip_addresses=None,
        affected_user_ids=None,
        affected_usernames=None,
        **kwargs
    ):
        """Retrieve audit logs, filtered based on given arguments.
        `Rest Documentation <https://developer.code42.com/api/#operation/rpc/search/search-audit-log>`__

        Args:
            begin_time (int or float or str or datetime, optional): Timestamp in milliseconds or
                str format "yyyy-MM-dd HH:MM:SS" or a datetime instance. Defaults to None.
            end_time (int or float or str or datetime, optional): Timestamp in milliseconds or
                str format "yyyy-MM-dd HH:MM:SS" or a datetime instance. Defaults to None.
            event_types (str or list, optional): A str or list of str of valid event types. Defaults to None.
            user_ids (str or list, optional): A str or list of str of Code42 userUids. Defaults to None.
            usernames (str or list, optional): A str or list of str of Code42 usernames. Defaults to None.
            user_ip_addresses (str or list, optional): A str or list of str of user ip addresses. Defaults to None.
            affected_user_ids (str or list, optional): A str or list of str of affected Code42 userUids. Defaults to None.
            affected_usernames (str or list, optional): A str or list of str of affected Code42 usernames. Defaults to None.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of audit logs.
        """
        return self._audit_log_service.get_all(
            begin_time=begin_time,
            end_time=end_time,
            event_types=event_types,
            user_ids=user_ids,
            usernames=usernames,
            user_ip_addresses=user_ip_addresses,
            affected_user_ids=affected_user_ids,
            affected_usernames=affected_usernames,
            **kwargs
        )



from py42.util import get_attribute_keys_from_class


class SortDirection(object):
    """Code42 request `sort_direction` constants for sorting returned lists in responses."""

    DESC = u"DESC"
    ASC = u"ASC"

    @staticmethod
    def choices():
        return get_attribute_keys_from_class(SortDirection)



from py42.exceptions import Py42InvalidRuleError
from py42.exceptions import Py42NotFoundError
from py42.services import BaseService


class AlertRulesService(BaseService):
    """A service to manage Alert Rules."""

    _api_prefix = u"/svc/api/v1/Rules/"

    def __init__(self, connection, user_context, user_profile_service):
        super(AlertRulesService, self).__init__(connection)
        self._user_context = user_context
        self._user_profile_service = user_profile_service
        self._exfiltration = None
        self._cloud_share = None
        self._file_type_mismatch = None

    @property
    def exfiltration(self):
        if not self._exfiltration:
            tenant_id = self._user_context.get_current_tenant_id()
            self._exfiltration = ExfiltrationService(self._connection, tenant_id)
        return self._exfiltration

    @property
    def cloudshare(self):
        if not self._cloud_share:
            tenant_id = self._user_context.get_current_tenant_id()
            self._cloud_share = CloudShareService(self._connection, tenant_id)
        return self._cloud_share

    @property
    def filetypemismatch(self):
        if not self._file_type_mismatch:
            tenant_id = self._user_context.get_current_tenant_id()
            self._file_type_mismatch = FileTypeMismatchService(
                self._connection, tenant_id
            )
        return self._file_type_mismatch

    def add_user(self, rule_id, user_id):
        tenant_id = self._user_context.get_current_tenant_id()
        user_details = self._user_profile_service.get_by_id(user_id)
        user_aliases = user_details.data.get(u"cloudUsernames") or []
        data = {
            u"tenantId": tenant_id,
            u"ruleId": rule_id,
            u"userList": [
                {u"userIdFromAuthority": user_id, u"userAliasList": user_aliases}
            ],
        }
        uri = u"{}{}".format(self._api_prefix, u"add-users")
        try:
            return self._connection.post(uri, json=data)
        except Py42NotFoundError as err:
            raise Py42InvalidRuleError(err, rule_id)

    def remove_user(self, rule_id, user_id):
        user_ids = [user_id]
        tenant_id = self._user_context.get_current_tenant_id()
        data = {u"tenantId": tenant_id, u"ruleId": rule_id, u"userIdList": user_ids}
        uri = u"{}{}".format(self._api_prefix, u"remove-users")
        return self._connection.post(uri, json=data)

    def remove_all_users(self, rule_id):
        tenant_id = self._user_context.get_current_tenant_id()
        data = {u"tenantId": tenant_id, u"ruleId": rule_id}
        uri = u"{}{}".format(self._api_prefix, u"remove-all-users")
        return self._connection.post(uri, json=data)


class CloudShareService(BaseService):
    _endpoint = u"/svc/api/v1/Rules/query-cloud-share-permissions-rule"

    def __init__(self, connection, tenant_id):
        super(CloudShareService, self).__init__(connection)
        self._tenant_id = tenant_id

    def get(self, rule_id):
        """Fetch cloud share alert rule by rule id.

        Args:
            rule_id (str): Observer rule Id of a rule to be fetched.

        Returns
            :class:`py42.response.Py42Response`
        """
        data = {u"tenantId": self._tenant_id, u"ruleIds": [rule_id]}
        return self._connection.post(self._endpoint, json=data)


class ExfiltrationService(BaseService):
    _endpoint = u"/svc/api/v1/Rules/query-endpoint-exfiltration-rule"

    def __init__(self, connection, tenant_id):
        super(ExfiltrationService, self).__init__(connection)
        self._tenant_id = tenant_id

    def get(self, rule_id):
        """Fetch exfiltration alert rule by rule id.

        Args:
            rule_id (str): Observer rule Id of a rule to be fetched.

        Returns
            :class:`py42.response.Py42Response`
        """
        data = {u"tenantId": self._tenant_id, u"ruleIds": [rule_id]}
        return self._connection.post(self._endpoint, json=data)


class FileTypeMismatchService(BaseService):
    _endpoint = u"/svc/api/v1/Rules/query-file-type-mismatch-rule"

    def __init__(self, connection, tenant_id):
        super(FileTypeMismatchService, self).__init__(connection)
        self._tenant_id = tenant_id

    def get(self, rule_id):
        """Fetch File type mismatch alert rules by rule id.

        Args:
            rule_id (str): Observer rule Id of a rule to be fetched.

        Returns
            :class:`py42.response.Py42Response`
        """
        data = {u"tenantId": self._tenant_id, u"ruleIds": [rule_id]}
        return self._connection.post(self._endpoint, json=data)



from py42.services import BaseService


class AdministrationService(BaseService):
    def get_diagnostics(self, include_volumes=None):
        uri = u"/api/Diagnostic"
        params = {u"incVolumes": include_volumes}
        return self._connection.get(uri, params=params)

    def get_alert_log(
        self, status=None, alert_type=None, page_num=None, page_size=None
    ):
        uri = u"/api/AlertLog"
        params = {
            u"status": status,
            u"type": alert_type,
            u"pgNum": page_num,
            u"pgSize": page_size,
        }
        return self._connection.get(uri, params=params)

    def get_current_tenant(self):
        uri = u"/c42api/v3/customer/my"
        return self._connection.get(uri)



from py42.exceptions import Py42BadRequestError
from py42.exceptions import Py42CaseAlreadyHasEventError
from py42.exceptions import Py42UpdateClosedCaseError
from py42.services import BaseService


class CasesFileEventsService(BaseService):

    _uri_prefix = u"/api/v1/case/{0}/fileevent"

    def __init__(self, connection):
        super(CasesFileEventsService, self).__init__(connection)

    def add(self, case_number, event_id):
        """Adds an event to the case.

        Args:
            case_number (int): Case number of the case.
            event_id (str): Event id to add to the case.

        Returns:
            :class:`py42.response.Py42Response`
        """
        try:
            return self._connection.post(
                u"{}/{}".format(self._uri_prefix.format(case_number), event_id)
            )
        except Py42BadRequestError as err:
            if "CASE_IS_CLOSED" in err.response.text:
                raise Py42UpdateClosedCaseError(err)
            elif "CASE_ALREADY_HAS_EVENT" in err.response.text:
                raise Py42CaseAlreadyHasEventError(err)
            else:
                raise

    def get(self, case_number, event_id):
        """Gets information of a specified event from the case.

        Args:
            case_number (int): Case number of the case.
            event_id (str): Event id to fetch from the case.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._connection.get(
            u"{}/{}".format(self._uri_prefix.format(case_number), event_id)
        )

    def get_all(self, case_number):
        """Gets all events associated with the given case.

        Args:
            case_number (int): Case number of the case.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._connection.get(self._uri_prefix.format(case_number))

    def delete(self, case_number, event_id):
        """Deletes an event from the case.

        Args:
            case_number (int): Case number of the case.
            event_id (str): Event id to remove from case.

        Returns:
            :class:`py42.response.Py42Response`
        """
        try:
            return self._connection.delete(
                u"{}/{}".format(self._uri_prefix.format(case_number), event_id)
            )
        except Py42BadRequestError as err:
            if "CASE_IS_CLOSED" in err.response.text:
                raise Py42UpdateClosedCaseError(err)
            raise



from py42._compat import quote
from py42.services import BaseService


class PreservationDataService(BaseService):
    def find_file_version(self, file_md5, file_sha256, paths):
        """Fetch file version details.

        Args:
            file_md5 (str): MD5 encoded hash of the file.
            file_sha256 (str): SHA256 encoded hash of the file.
            paths (str): File path with filename to fetch.

        Returns:
            :class:`py42.response.Py42Response`
        """

        data = {u"fileSHA256": file_sha256, u"fileMD5": file_md5, u"devicePaths": paths}
        uri = u"/api/v1/FindAvailableVersion"
        return self._connection.post(uri, json=data)

    def get_file_version_list(self, deviceGuid, file_md5, file_sha256, path):
        params = u"fileSHA256={}&fileMD5={}&deviceGuid={}&path={}"
        params = params.format(file_sha256, file_md5, deviceGuid, quote(path))
        uri = u"/api/v1/FileVersionListing?{}".format(params)
        return self._connection.get(uri)



from py42.clients._archiveaccess import ArchiveContentPusher
from py42.clients._archiveaccess.restoremanager import create_file_size_poller
from py42.clients._archiveaccess.restoremanager import create_restore_job_manager


class ArchiveAccessorFactory(object):
    """Creates different types of :class:`py42.clients._archiveaccess.ArchiveAccessor`
    for use in a web/push restore."""

    def __init__(self, archive_service, storage_service_factory):
        self._archive_service = archive_service
        self._storage_service_factory = storage_service_factory

    def create_archive_accessor(
        self,
        device_guid,
        accessor_class,
        destination_guid=None,
        private_password=None,
        encryption_key=None,
    ):
        """Used for creating instances of ArchiveExplorer or ArchiveContentStreamer.
        Web restore uses this method to create an ArchiveContentStreamer and push
        restore uses create_archive_content_pusher() to create an ArchiveContentPusher
        and this method to create an ArchiveExplorer."""
        destination_guid = (
            destination_guid
            or self._storage_service_factory.auto_select_destination_guid(device_guid)
        )
        storage_archive_service = self._storage_service_factory.create_archive_service(
            device_guid, destination_guid
        )
        (
            decryption_keys,
            session_id,
            restore_job_manager,
            file_size_poller,
        ) = self._create_archive_accessor_dependencies(
            storage_archive_service, device_guid, private_password, encryption_key,
        )
        return accessor_class(
            device_guid,
            session_id,
            destination_guid,
            storage_archive_service,
            restore_job_manager,
            file_size_poller,
        )

    def create_archive_content_pusher(
        self,
        device_guid,
        accepting_guid,
        destination_guid=None,
        private_password=None,
        encryption_key=None,
    ):
        """Creates a class that inherits from the ArchiveAccessor class that has method
        stream_to_device()."""
        push_service = self._storage_service_factory.create_push_restore_service(
            accepting_guid
        )
        (
            decryption_keys,
            session_id,
            restore_job_manager,
            file_size_poller,
        ) = self._create_archive_accessor_dependencies(
            push_service, device_guid, private_password, encryption_key,
        )
        destination_guid = (
            destination_guid
            or self._storage_service_factory.auto_select_destination_guid(device_guid)
        )
        node_guid = self._get_node_guid(device_guid, destination_guid)
        return ArchiveContentPusher(
            device_guid,
            destination_guid,
            node_guid,
            session_id,
            push_service,
            restore_job_manager,
            file_size_poller,
        )

    def _create_archive_accessor_dependencies(
        self, storage_archive_service, device_guid, private_password, encryption_key
    ):
        decryption_keys = self._get_decryption_keys(
            device_guid, private_password, encryption_key,
        )
        session_id = self._create_restore_session(
            storage_archive_service, device_guid, **decryption_keys
        )
        restore_job_manager = create_restore_job_manager(
            storage_archive_service, device_guid, session_id,
        )
        file_size_poller = create_file_size_poller(storage_archive_service, device_guid)
        return decryption_keys, session_id, restore_job_manager, file_size_poller

    def _get_decryption_keys(self, device_guid, private_password, encryption_key):
        decryption_keys = {}
        # Favors encryption-key over other security levels.
        if encryption_key:
            decryption_keys[u"encryption_key"] = encryption_key
        else:
            data_key_token = (
                self._get_data_key_token(device_guid) if not encryption_key else None
            )
            if data_key_token:
                decryption_keys[u"data_key_token"] = data_key_token
            if private_password:
                decryption_keys[u"private_password"] = private_password
        return decryption_keys

    def _get_data_key_token(self, device_guid):
        return self._archive_service.get_data_key_token(device_guid)[u"dataKeyToken"]

    def _get_node_guid(self, device_guid, destination_guid):
        response = self._archive_service.get_web_restore_info(
            device_guid, destination_guid
        )
        return response[u"nodeGuid"]

    @staticmethod
    def _create_restore_session(session_creator, device_guid, **kwargs):
        """Session creator is :class:`StorageArchiveService` for web restore
        and :class:`PushRestoreService` for push restore."""
        response = session_creator.create_restore_session(device_guid, **kwargs)
        return response[u"webRestoreSessionId"]



import posixpath
from collections import namedtuple

from py42.exceptions import Py42ArchiveFileNotFoundError

# Data for initiating a web or push restore.
FileSelection = namedtuple(u"FileSelection", u"file, num_files, num_dirs, num_bytes")


class FileType(object):
    """The different file-types in an archive."""

    DIRECTORY = u"DIRECTORY"
    FILE = u"FILE"


class ArchiveAccessor(object):
    """Base class for certain archive operations, such file-exploring or restoring."""

    DEFAULT_DIRECTORY_DOWNLOAD_NAME = u"download"
    JOB_POLLING_INTERVAL = 1

    def __init__(
        self,
        device_guid,
        archive_session_id,
        destination_guid,
        storage_archive_service,
        restore_job_manager,
        file_size_poller,
    ):
        self._device_guid = device_guid
        self._archive_session_id = archive_session_id
        self.destination_guid = destination_guid
        self._storage_archive_service = storage_archive_service
        self._restore_job_manager = restore_job_manager
        self._file_size_poller = file_size_poller


class ArchiveExplorer(ArchiveAccessor):
    """Abstracts file exploring / tree-navigation in the context of a restore."""

    def create_file_selections(self, backup_set_id, file_paths, file_size_calc_timeout):
        if not isinstance(file_paths, (list, tuple)):
            file_paths = [file_paths]
        file_paths = [fp.replace(u"\\", u"/") for fp in file_paths]
        metadata_list = self._get_restore_metadata(backup_set_id, file_paths)
        file_ids = [md[u"id"] for md in metadata_list]
        file_sizes = self._file_size_poller.get_file_sizes(
            file_ids, timeout=file_size_calc_timeout
        )
        return _create_file_selections(file_paths, metadata_list, file_sizes)

    def _get_restore_metadata(self, backup_set_id, file_paths):
        metadata_list = []
        for path in file_paths:
            metadata = self._get_file_via_walking_tree(backup_set_id, path)
            metadata_list_entry = {
                u"id": metadata[u"id"],
                u"path": metadata[u"path"],
                u"type": metadata[u"type"],
            }
            metadata_list.append(metadata_list_entry)
        return metadata_list

    def _get_file_via_walking_tree(self, backup_set_id, file_path):
        path_parts = file_path.split(u"/")
        path_root = path_parts[0] + u"/"

        response = self._get_children(backup_set_id, file_id=None)
        for root in response:
            if root[u"path"].lower() == path_root.lower():
                return self._walk_tree(backup_set_id, response, root, path_parts[1:])

        raise Py42ArchiveFileNotFoundError(response, self._device_guid, file_path)

    def _walk_tree(
        self, backup_set_id, response, current_file, remaining_path_components
    ):
        if not remaining_path_components or not remaining_path_components[0]:
            return current_file

        children = self._get_children(backup_set_id, file_id=current_file[u"id"])
        current_path = current_file[u"path"]
        target_child_path = posixpath.join(current_path, remaining_path_components[0])

        for child in children:
            if child[u"path"].lower() == target_child_path.lower():
                return self._walk_tree(
                    backup_set_id, response, child, remaining_path_components[1:]
                )

        raise Py42ArchiveFileNotFoundError(
            response, self._device_guid, target_child_path
        )

    def _get_children(self, backup_set_id, file_id=None):
        return self._storage_archive_service.get_file_path_metadata(
            self._archive_session_id,
            self._device_guid,
            backup_set_id,
            file_id=file_id,
            show_deleted=True,
        )


class ArchiveContentStreamer(ArchiveExplorer):
    """A class with methods for restoring files from backup."""

    def stream_from_backup(
        self, backup_set_id, file_paths, file_size_calc_timeout=None, show_deleted=None,
    ):
        file_selections = self.create_file_selections(
            backup_set_id, file_paths, file_size_calc_timeout
        )
        return self._restore_job_manager.get_stream(
            backup_set_id, file_selections, show_deleted=show_deleted
        )


class ArchiveContentPusher(ArchiveAccessor):
    """A class with methods for restoring files from backup and pushing them to a device
    (push restore)."""

    def __init__(
        self,
        device_guid,
        destination_guid,
        node_guid,
        archive_session_id,
        storage_archive_service,
        restore_job_manager,
        file_size_poller,
    ):
        self._node_guid = node_guid
        super(ArchiveContentPusher, self).__init__(
            device_guid,
            archive_session_id,
            destination_guid,
            storage_archive_service,
            restore_job_manager,
            file_size_poller,
        )

    def stream_to_device(
        self,
        restore_path,
        accepting_guid,
        file_selections,
        backup_set_id,
        show_deleted,
        overwrite_existing_files,
    ):
        return self._restore_job_manager.send_stream(
            restore_path,
            self._node_guid,
            accepting_guid,
            file_selections,
            backup_set_id,
            show_deleted,
            overwrite_existing_files,
        )


def _create_file_selections(file_paths, metadata_list, file_sizes=None):
    file_selections = []
    for i in range(0, len(file_paths)):
        metadata = metadata_list[i]
        size_info = file_sizes[i] if file_sizes else _get_default_file_size()
        file = {
            u"fileType": metadata[u"type"].upper(),
            u"path": metadata[u"path"],
            u"selected": True,
        }
        selection = FileSelection(
            file, size_info[u"numFiles"], size_info[u"numDirs"], size_info[u"size"],
        )
        file_selections.append(selection)
    return file_selections


def _get_default_file_size():
    return {u"numFiles": 1, u"numDirs": 1, u"size": 1}



from py42.sdk.queries.alerts.filters import AlertState


class AlertsClient(object):
    """A client to expose alert API.

    `Rest Documentation <https://developer.code42.com/api/#tag/Alerts>`__
    """

    def __init__(self, alert_service, alert_rules_client):
        self._alert_service = alert_service
        self._alert_rules_client = alert_rules_client

    @property
    def rules(self):
        """A collection of methods for managing alert rules.

        Returns:
            :class:`py42.services.alertrules.AlertRulesClient`
        """
        return self._alert_rules_client

    def search(self, query, page_num=1, page_size=None):
        """Searches alerts using the given :class:`py42.sdk.queries.alerts.alert_query.AlertQuery`.

        `Rest Documentation <https://developer.code42.com/api/#operation/Alerts_QueryAlert>`__

        Args:
            query (:class:`py42.sdk.queries.alerts.alert_query.AlertQuery`): An alert query.
                See the :ref:`Executing Searches User Guide <anchor_search_alerts>` to learn more
                about how to construct a query.
            page_num (int, optional): The page number to get. Defaults to 1.
            page_size (int, optional): The number of items per page. Defaults to `py42.settings.items_per_page`.


        Returns:
            :class:`py42.response.Py42Response`: A response containing the alerts that match the given
            query.
        """
        return self._alert_service.search(query, page_num, page_size)

    def search_all_pages(self, query):
        """Searches alerts using the given :class:`py42.sdk.queries.alerts.alert_query.AlertQuery`.

        `Rest Documentation <https://developer.code42.com/api/#operation/Alerts_QueryAlert>`__

        Args:
            query (:class:`py42.sdk.queries.alerts.alert_query.AlertQuery`): An alert query.
                See the :ref:`Executing Searches User Guide <anchor_search_alerts>` to learn more
                about how to construct a query.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of alerts that match the given query.
        """
        return self._alert_service.search_all_pages(query)

    def get_details(self, alert_ids):
        """Gets the details for the alerts with the given IDs, including the file event query that,
        when passed into a search, would result in events that could have triggered the alerts.

        `Rest Documentation <https://developer.code42.com/api/#operation/Alerts_QueryAlertDetails>`__

        Args:
            alert_ids (str or list[str]): The identification number(s) of the alerts for which you want to
                get details for. Note: The alerts backend accepts a maximum of 100 alerts per request.

        Returns:
            :class:`py42.response.Py42Response`: A response containing the alert details.
        """
        return self._alert_service.get_details(alert_ids)

    def resolve(self, alert_ids, reason=None):
        """Resolves the alerts with the given IDs.

        Args:
            alert_ids (str or list[str]): The identification number(s) for the alerts to resolve.
                Note: The alerts backend accepts a maximum of 100 alerts per request.
            reason (str, optional): The reason the alerts are now resolved. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._alert_service.update_state(
            AlertState.DISMISSED, alert_ids, note=reason
        )

    def reopen(self, alert_ids, reason=None):
        """Reopens the resolved alerts with the given IDs.

        Args:
            alert_ids (str or list[str]): The identification number(s) for the alerts to reopen.
                Note: The alerts backend accepts a maximum of 100 alerts per request.
            reason (str, optional): The reason the alerts are reopened. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._alert_service.update_state(AlertState.OPEN, alert_ids, note=reason)

    def update_state(self, status, alert_ids, note=None):
        """Update status for given alert IDs.

        Args:
            status (str): Status to set from OPEN, RESOLVED, PENDING, IN_PROGRESS
            alert_ids (str or list[str]): The identification number(s) for the alerts to reopen.
                Note: The alerts backend accepts a maximum of 100 alerts per request.
            note (str, optional): User note regarding the status. Defaults to None.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._alert_service.update_state(status, alert_ids, note=note)

    def update_note(self, alert_id, note):
        """Add/update a note to an alert.

        Args:
            alert_id (str): The identification number of an alert to add a note to.
            note (str): User note regarding the alert, in less than 2000 characters.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._alert_service.update_note(alert_id, note)

    def get_aggregate_data(self, alert_id):
        """Gets alert summary with details about observations.

        Args:
            alert_id (str): Gets the details for the alert with the given ID.

        Returns:
            :class:`py42.response.Py42Response`
        """
        return self._alert_service.get_aggregate_data(alert_id)



from py42 import settings
from py42.exceptions import Py42BadRequestError
from py42.exceptions import Py42CaseNameExistsError
from py42.exceptions import Py42DescriptionLimitExceededError
from py42.exceptions import Py42InvalidCaseUserError
from py42.exceptions import Py42UpdateClosedCaseError
from py42.services import BaseService
from py42.services.util import get_all_pages


class CasesService(BaseService):

    _uri_prefix = u"/api/v1/case"

    def __init__(self, connection):
        super(CasesService, self).__init__(connection)

    def create(
        self, name, subject=None, assignee=None, description=None, findings=None
    ):
        data = {
            u"assignee": assignee,
            u"description": description,
            u"findings": findings,
            u"name": name,
            u"subject": subject,
        }
        try:
            return self._connection.post(self._uri_prefix, json=data)
        except Py42BadRequestError as err:
            _handle_common_invalid_case_parameters_errors(err, name)

    def get_page(
        self,
        page_num,
        name=None,
        status=None,
        created_at=None,
        updated_at=None,
        subject=None,
        assignee=None,
        page_size=None,
        sort_direction=u"asc",
        sort_key=u"number",
        **kwargs
    ):

        page_size = page_size or settings.items_per_page
        params = {
            u"name": name,
            u"subject": subject,
            u"assignee": assignee,
            u"createdAt": created_at,
            u"updatedAt": updated_at,
            u"status": status,
            u"pgNum": page_num,
            u"pgSize": page_size,
            u"srtDir": sort_direction,
            u"srtKey": sort_key,
        }
        params.update(**kwargs)

        return self._connection.get(self._uri_prefix, params=params)

    def get_all(
        self,
        name=None,
        status=None,
        created_at=None,
        updated_at=None,
        subject=None,
        assignee=None,
        page_size=None,
        sort_direction=u"asc",
        sort_key=u"number",
        **kwargs
    ):
        return get_all_pages(
            self.get_page,
            u"cases",
            name=name,
            status=status,
            created_at=created_at,
            updated_at=updated_at,
            assignee=assignee,
            subject=subject,
            page_size=page_size,
            sort_direction=sort_direction,
            sort_key=sort_key,
            **kwargs
        )

    def get(self, case_number):
        return self._connection.get("{}/{}".format(self._uri_prefix, case_number))

    def export_summary(self, case_number):
        uri_prefix = u"{}/{}/{}".format(self._uri_prefix, case_number, u"export")
        return self._connection.get(uri_prefix)

    def update(
        self,
        case_number,
        name=None,
        subject=None,
        assignee=None,
        description=None,
        findings=None,
        status=None,
    ):
        current_case_data = self.get(case_number).data

        data = {
            u"assignee": assignee or current_case_data.get(u"assignee"),
            u"description": description or current_case_data.get(u"description"),
            u"findings": findings or current_case_data.get(u"findings"),
            u"name": name or current_case_data.get(u"name"),
            u"subject": subject or current_case_data.get(u"subject"),
            u"status": status or current_case_data.get("status"),
        }
        try:
            return self._connection.put(
                u"{}/{}".format(self._uri_prefix, case_number), json=data
            )
        except Py42BadRequestError as err:
            if u"NO_EDITS_ONCE_CLOSED" in err.response.text:
                raise Py42UpdateClosedCaseError(err)
            _handle_common_invalid_case_parameters_errors(err, name)


def _handle_common_invalid_case_parameters_errors(base_err, name):
    if u"NAME_EXISTS" in base_err.response.text:
        raise Py42CaseNameExistsError(base_err, name)
    elif u"NO_EDITS_ONCE_CLOSED" in base_err.response.text:
        raise Py42UpdateClosedCaseError(base_err)
    elif u"DESCRIPTION_TOO_LONG" in base_err.response.text:
        raise Py42DescriptionLimitExceededError(base_err)
    elif u"INVALID_USER" in base_err.response.text:
        if u"subject" in base_err.response.text:
            raise Py42InvalidCaseUserError(base_err, u"subject")
        elif u"assignee" in base_err.response.text:
            raise Py42InvalidCaseUserError(base_err, u"assignee")
    raise



import time

from py42.services.storage.restore import PushRestoreExistingFiles
from py42.services.storage.restore import PushRestoreLocation
from py42.settings import debug
from py42.util import format_dict


def create_restore_job_manager(
    storage_archive_service, device_guid, archive_session_id
):
    return RestoreJobManager(storage_archive_service, device_guid, archive_session_id,)


def create_file_size_poller(storage_archive_service, device_guid):
    return FileSizePoller(storage_archive_service, device_guid)


class _RestorePoller(object):
    """Base class for observing web/push restore calculation processes, such as polling
    for file sizes or the restore itself."""

    JOB_POLLING_INTERVAL_SECONDS = 1

    def __init__(self, storage_archive_service, device_guid, job_polling_interval=None):
        self._storage_archive_service = storage_archive_service
        self._device_guid = device_guid
        self._job_polling_interval = (
            job_polling_interval or self.JOB_POLLING_INTERVAL_SECONDS
        )


class FileSizePoller(_RestorePoller):
    """Monitors the status of a poll-job; the bytes and number of files needed for a
    restore. This affords py42 users a chance to observe the progress of a web/push
    restore."""

    def __init__(self, storage_archive_service, device_guid, job_polling_interval=None):
        super(FileSizePoller, self).__init__(
            storage_archive_service, device_guid, job_polling_interval
        )

    def get_file_sizes(self, file_ids, timeout):
        if not timeout:
            # Skips file size calculation
            return None

        job_ids = self._start_poll(file_ids)
        return self._wait_for_jobs(job_ids, timeout)

    def _start_poll(self, file_ids):
        job_ids = []
        for file_id in file_ids:
            response = self._storage_archive_service.create_file_size_job(
                self._device_guid, file_id
            )
            job_id = response[u"jobId"]
            job_ids.append(job_id)
        return job_ids

    def _wait_for_jobs(self, job_ids, timeout):
        t0 = time.time()
        sizes = []

        # Waits until the job_ids stack is empty
        while job_ids:
            for job_id in job_ids:
                response = self._get_job_status(job_id)
                size_dict = _create_size_dict(job_id, response)
                _print_file_size(size_dict)
                if response[u"status"].lower() == u"done":
                    job_ids.remove(job_id)
                    sizes.append(size_dict)

            # File size calculation is taking too long.
            if time.time() - t0 > timeout:
                return None

            if job_ids:
                time.sleep(self._job_polling_interval)
        return sizes

    def _get_job_status(self, job_id):
        return self._storage_archive_service.get_file_size_job(
            job_id, self._device_guid
        )


class RestoreJobManager(_RestorePoller):
    """Monitors and manages the status of a web/push restore."""

    def __init__(
        self,
        storage_archive_service,
        device_guid,
        archive_session_id,
        job_polling_interval=None,
    ):
        super(RestoreJobManager, self).__init__(
            storage_archive_service=storage_archive_service,
            device_guid=device_guid,
            job_polling_interval=job_polling_interval,
        )
        self._archive_session_id = archive_session_id

    def get_stream(self, backup_set_id, file_selections, show_deleted):
        response = self._start_web_restore(backup_set_id, file_selections, show_deleted)
        job_id = self._wait_for_job(response)
        return self._get_stream(job_id)

    def send_stream(
        self,
        restore_path,
        node_guid,
        accepting_guid,
        file_selections,
        backup_set_id,
        show_deleted,
        overwrite_existing_files,
    ):
        num_files = sum([fs.num_files for fs in file_selections])
        num_bytes = sum([fs.num_bytes for fs in file_selections])
        file_location = None
        permit_restore_to_different_os_version = False
        existing_files = PushRestoreExistingFiles.RENAME_ORIGINAL

        # Use expected request parameters for restoring to the original location.
        if restore_path == PushRestoreLocation.ORIGINAL_LOCATION:
            file_location = restore_path
            restore_path = u""
            if self._device_guid != accepting_guid:
                permit_restore_to_different_os_version = True
            if overwrite_existing_files:
                existing_files = PushRestoreExistingFiles.OVERWRITE_ORIGINAL

        return self._storage_archive_service.start_push_restore(
            self._device_guid,
            accepting_guid,
            self._archive_session_id,
            node_guid,
            restore_path,
            [
                {
                    u"backupSetId": backup_set_id,
                    u"files": [f.file for f in file_selections],
                }
            ],
            num_files,
            num_bytes,
            show_deleted=show_deleted,
            file_location=file_location,
            permit_restore_to_different_os_version=permit_restore_to_different_os_version,
            existing_files=existing_files,
        )

    def _wait_for_job(self, response):
        job_id = response[u"jobId"]
        while not self._is_job_complete(job_id):
            time.sleep(self._job_polling_interval)
        return job_id

    def _is_job_complete(self, job_id):
        response = self._storage_archive_service.get_restore_status(job_id)
        is_done = response[u"done"]
        percentage_dict = {
            u"jobId": job_id,
            u"status": response.data.get(u"status"),
            u"percentComplete": response[u"percentComplete"] if not is_done else 100,
        }
        debug.logger.debug(format_dict(percentage_dict))
        return is_done

    def _start_web_restore(self, backup_set_id, file_selections, show_deleted):
        num_files = sum([fs.num_files for fs in file_selections])
        num_dirs = sum([fs.num_dirs for fs in file_selections])
        num_bytes = sum([fs.num_bytes for fs in file_selections])

        # For py42 backwards compat.
        if show_deleted is None:
            show_deleted = True

        return self._storage_archive_service.start_restore(
            self._device_guid,
            self._archive_session_id,
            [
                {
                    u"backupSetId": backup_set_id,
                    u"files": [f.file for f in file_selections],
                }
            ],
            num_files,
            num_dirs,
            num_bytes,
            show_deleted=show_deleted,
        )

    def _get_stream(self, job_id):
        response = self._storage_archive_service.stream_restore_result(job_id)
        return response


def _create_size_dict(job_id, size_response):
    size_dict = size_response.data
    size_dict[u"jobId"] = job_id
    return size_dict


def _print_file_size(size_dict):
    debug.logger.debug(format_dict(size_dict))



from py42.exceptions import Py42BadRequestError
from py42.exceptions import Py42NotFoundError
from py42.exceptions import Py42UserNotOnListError
from py42.services import BaseService
from py42.services.detectionlists import _DetectionListFilters
from py42.services.detectionlists import _PAGE_SIZE
from py42.services.detectionlists import handle_user_already_added_error
from py42.services.util import get_all_pages
from py42.util import get_attribute_keys_from_class


class HighRiskEmployeeFilters(_DetectionListFilters):
    """Constants available for filtering High Risk Employee search results."""

    @staticmethod
    def choices():
        return get_attribute_keys_from_class(HighRiskEmployeeFilters)


class HighRiskEmployeeService(BaseService):
    """A service for interacting with High Risk Employee APIs."""

    _resource = u"v2/highriskemployee"

    def __init__(self, connection, user_context, user_profile_service):
        super(HighRiskEmployeeService, self).__init__(connection)
        self._user_context = user_context
        self._user_profile_service = user_profile_service

    def _make_uri(self, action):
        return u"{}{}".format(self._resource, action)

    def _add_high_risk_employee(self, tenant_id, user_id):
        data = {u"tenantId": tenant_id, u"userId": user_id}
        uri = self._make_uri(u"/add")
        return self._connection.post(uri, json=data)

    def add(self, user_id):
        """Adds a user to the High Risk Employee detection list.

        Raises a :class:`Py42UserAlreadyAddedError` when a user already exists in the High Risk Employee
        detection list.
        `REST Documentation <https://developer.code42.com/api/#operation/HighRiskEmployeeControllerV2_AddEmployee>`__

        Args:
            user_id (str or int): The Code42 userUid of the user you want to add to the High Risk
                Employee detection list.

        Returns:
            :class:`py42.response.Py42Response`
        """
        tenant_id = self._user_context.get_current_tenant_id()
        try:
            return self._add_high_risk_employee(tenant_id, user_id)
        except Py42BadRequestError as err:
            handle_user_already_added_error(err, user_id, u"high-risk-employee list")
            raise

    def set_alerts_enabled(self, enabled=True):
        """Enables alerts.
        `Rest Documentation <https://developer.code42.com/api/#operation/HighRiskEmployeeControllerV2_SetAlertState>`__

        Args:
            enabled (bool): Whether to enable alerts for all users.

        Returns:
            :class:`py42.response.Py42Response`
        """
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"alertsEnabled": enabled,
        }
        uri = self._make_uri(u"/setalertstate")
        return self._connection.post(uri, json=data)

    def remove(self, user_id):
        """Removes a user from the High Risk Employee detection list.
        `Rest Documentation <https://developer.code42.com/api/#operation/HighRiskEmployeeControllerV2_RemoveUser>`__

        Args:
            user_id (str or int): The Code42 userUid of the user you want to add to the High Risk
                Employee detection list.

        Returns:
            :class:`py42.response.Py42Response`
        """
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"userId": user_id,
        }
        uri = self._make_uri(u"/remove")
        try:
            return self._connection.post(uri, json=data)
        except Py42NotFoundError as err:
            raise Py42UserNotOnListError(err, user_id, u"high-risk-employee")

    def get(self, user_id):
        """Gets user information.
        `Rest Documentation <https://developer.code42.com/api/#operation/HighRiskEmployeeControllerV2_GetEmployee>`__

        Args:
            user_id (str or int): The Code42 userUid of the user has been added to the High Risk
                Employee detection list.

        Returns:
            :class:`py42.response.Py42Response`
        """
        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"userId": user_id,
        }
        uri = self._make_uri(u"/get")
        return self._connection.post(uri, json=data)

    def get_all(
        self,
        filter_type=HighRiskEmployeeFilters.OPEN,
        sort_key=None,
        sort_direction=None,
        page_size=_PAGE_SIZE,
    ):
        """Searches High Risk Employee list. Filter results by filter_type.
        `Rest Documentation <https://developer.code42.com/api/#operation/HighRiskEmployeeControllerV2_Search>`__

        Args:
            filter_type (str, optional): ``EXFILTRATION_30_DAYS``, ``EXFILTRATION_24_HOURS``,
                or ``OPEN``. Constants are available at
                :class:`py42.services.detectionlists.high_risk_employee.HighRiskEmployeeFilters`.
                Defaults to "OPEN".
            sort_key (str, optional): Sort results based by field. Defaults to None.
            sort_direction (str, optional): ``ASC`` or ``DESC``. Constants available at
                :class:`py42.constants.SortDirection`. Defaults to None.
            page_size (int, optional): The number of high risk employees to return
                per page. Defaults to 100.

        Returns:
            generator: An object that iterates over :class:`py42.response.Py42Response` objects
            that each contain a page of users.
        """

        return get_all_pages(
            self.get_page,
            u"items",
            filter_type=filter_type,
            sort_key=sort_key,
            sort_direction=sort_direction,
            page_size=page_size or _PAGE_SIZE,
        )

    def get_page(
        self,
        page_num,
        filter_type=HighRiskEmployeeFilters.OPEN,
        sort_key=None,
        sort_direction=None,
        page_size=_PAGE_SIZE,
    ):
        """Gets a single page of High Risk Employees.

        Args:
            page_num (int): The page number to request.
            filter_type (str, optional): ``EXFILTRATION_30_DAYS``, ``EXFILTRATION_24_HOURS``,
                or ``OPEN``. Constants are available at
                :class:`py42.services.detectionlists.high_risk_employee.HighRiskEmployeeFilters`.
                Defaults to "OPEN".
            sort_key (str, optional): Sort results based by field. Defaults to None.
            sort_direction (str. optional): ``ASC`` or ``DESC``. Constants available at
                :class:`py42.constants.SortDirection`. Defaults to None.
            page_size (int, optional): The number of high risk employees to return
                per page. Defaults to 100.

        Returns:
            :class:`py42.response.Py42Response`
        """

        data = {
            u"tenantId": self._user_context.get_current_tenant_id(),
            u"filterType": filter_type,
            u"pgNum": page_num,
            u"pgSize": page_size,
            u"srtKey": sort_key,
            u"srtDirection": sort_direction,
        }
        uri = self._make_uri(u"/search")
        return self._connection.post(uri, json=data)
